# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for language model."""

import os
import tempfile
from typing import Type
import unittest

import langfun.core as lf
from langfun.core.eval import base
from langfun.core.llms import fake
import langfun.core.structured as lf_structured
import pyglove as pg


# We put class definitions outside the functors just to make it easier
# to refer to them in test.
class Solution(pg.Object):
  final_answer: int


class SolutionForCompletion(pg.Object):
  question: str
  final_answer: int


@pg.functor
def answer_schema():
  return Solution


@pg.functor
def answer_schema_with_fewshot_examples(evaluation):
  del evaluation
  return Solution, [
      lf_structured.MappingExample(
          input='The result of one plus two',
          output=Solution(3),
          schema=Solution,
      )
  ]


@pg.functor
def complete_schema():
  return SolutionForCompletion


def eval_set(
    eval_id: str,
    method: str,
    schema_fn,
    lm: lf.LanguageModel = pg.MISSING_VALUE,
    use_cache: bool = True,
    cls: Type[base.Evaluation] = base.Evaluation,
    **kwargs,
):
  """Creates an evaluation object for testing."""
  tmp_dir = tempfile.gettempdir()
  return cls(
      root_dir=os.path.join(tmp_dir, eval_id),
      inputs=base.as_inputs([
          pg.Dict(question='Compute 1 + 1'),
          pg.Dict(question='Compute 1 + 2'),
      ]),
      method=method,
      prompt='{{example.question}}',
      completion_prompt_field='question',
      schema_fn=schema_fn,
      lm=lm,
      use_cache=use_cache,
      max_workers=1,
      **kwargs
  )


class EvaluationTest(unittest.TestCase):
  """Evaluation test."""

  def setUp(self):
    super().setUp()
    pg.symbolic.set_save_handler(pg.symbolic.default_save_handler)
    pg.symbolic.set_load_handler(pg.symbolic.default_load_handler)

  def test_basics(self):
    lm = fake.StaticSequence(['two', 'Solution(final_answer=2)'])
    s = eval_set('basic_test', 'call', schema_fn=answer_schema(), lm=lm)

    self.assertEqual(s.dir, os.path.join(s.root_dir, s.id))
    self.assertEqual(s.hash, s.clone().hash)
    # Test persistent hash.
    self.assertEqual(s.hash, 'e43392e4')
    self.assertEqual(
        s.hash, s.clone(override={'max_workers': 2, 'lm.timeout': 20}).hash
    )
    self.assertNotEqual(
        s.hash, s.clone(override={'prompt': 'Hello {{example.question}}'}).hash
    )
    self.assertIsNone(s.parent)
    self.assertIs(s.schema.spec.cls, Solution)
    self.assertIsNone(s.fewshot_examples)

    # Test schema_fn with fewshot examples.
    s.rebind(schema_fn=answer_schema_with_fewshot_examples())
    self.assertIs(s.schema.spec.cls, Solution)
    self.assertTrue(
        pg.eq(
            s.fewshot_examples,
            [
                lf_structured.MappingExample(
                    input='The result of one plus two',
                    output=Solution(3),
                    schema=Solution,
                )
            ],
        )
    )

  def test_schema_for_completion(self):

    @pg.functor()
    def _answer_schema():

      class Solution1(pg.Object):
        final_answer: int

      return Solution1, [
          lf.structured.MappingExample(
              input='The result of one plus two',
              output=Solution1(3),
              schema=Solution1,
          )
      ]

    s = eval_set(
        'schema_for_completion', 'complete',
        schema_fn=_answer_schema(), lm=fake.StaticResponse('hi'))

    fewshot_examples = s.fewshot_examples
    solution_cls = s.schema.spec.cls

    # Verify class schema get updated.
    self.assertEqual('question', list(solution_cls.__schema__.keys())[0])

    # Verify query examples are mapped to completion examples.
    self.assertTrue(
        pg.eq(
            fewshot_examples,
            [
                lf.structured.MappingExample(
                    input=solution_cls.partial(
                        question='The result of one plus two'
                    ),
                    output=solution_cls('The result of one plus two', 3),
                )
            ],
        )
    )

  def test_bad_init(self):

    @pg.functor()
    def _bad_completion_schema():
      return int

    s = eval_set(
        'bad_init2', 'complete',
        schema_fn=_bad_completion_schema(), lm=fake.StaticResponse('hi'))

    with self.assertRaisesRegex(TypeError, '.*must be .*class.*'):
      _ = s.schema

  def test_dryrun(self):
    lm = fake.StaticResponse('Solution(final_answer=2)')
    s = eval_set('dryrun_test', 'query', schema_fn=answer_schema(), lm=lm)
    s.dryrun(verbose=True)
    self.assertEqual(
        s.dryrun_output,
        lf.AIMessage(
            text='Solution(final_answer=2)',
            result=Solution(2),
            cache_seed=0,
            score=1.0,
            logprobs=None,
            is_cached=False,
            usage=lf.LMSamplingUsage(428, 24, 452),
            tags=['lm-response', 'lm-output', 'transformed'],
        ),
    )

  def test_run(self):
    lm = fake.StaticSequence([
        'Solution(final_answer=2)',
        '3',
    ])
    s = eval_set('run_test', 'query', schema_fn=answer_schema(), lm=lm)
    s.run()
    self.assertEqual(
        s.result,
        dict(
            experiment_setup=dict(
                id='Evaluation@2fbf1b05',
                dir=s.dir,
                model='StaticSequence',
                prompt_template='{{example.question}}',
                method='query',
                schema_fn='answer_schema()',
            ),
            cache_stats=dict(
                use_cache=True, num_queries=2, num_hits=0, num_updates=2
            ),
            metrics=dict(
                total=2,
                failures=1,
                failure_rate=0.5,
                oop_failures=1,
                oop_failure_rate=0.5,
                non_oop_failures=0,
                non_oop_failure_rate=0.0,
                failure_breakdown={
                    'MappingError.SchemaError.TypeError': 1
                }
            ),
            usage=dict(
                total_prompt_tokens=856,
                total_completion_tokens=25,
                num_usages=2,
                average_prompt_tokens=428,
                average_completion_tokens=12,
                average_total_tokens=440,
            ),
        ),
    )
    self.assertTrue(
        os.path.exists(os.path.join(s.dir, base.Evaluation.EXPERIMENT_JSON)))
    self.assertTrue(
        os.path.exists(os.path.join(s.dir, base.Evaluation.RESULT_JSON)))
    self.assertTrue(
        os.path.exists(os.path.join(s.dir, base.Evaluation.OOP_FAILURES_JSON)))
    self.assertTrue(
        os.path.exists(
            os.path.join(s.dir, base.Evaluation.NON_OOP_FAILURES_JSON)))
    self.assertTrue(
        os.path.exists(os.path.join(s.dir, base.Evaluation.CACHE_JSON)))
    self.assertTrue(
        os.path.exists(os.path.join(s.dir, base.Evaluation.INDEX_HTML)))
    self.assertTrue(
        os.path.exists(os.path.join(s.dir, base.Evaluation.OOP_FAILURES_HTML)))
    self.assertTrue(
        os.path.exists(
            os.path.join(s.dir, base.Evaluation.NON_OOP_FAILURES_HTML)))
    self.assertTrue(
        os.path.exists(os.path.join(s.root_dir, base.Evaluation.SUMMARY_HTML))
    )
    # Check summary JSON.
    summary_json = os.path.join(
        s.root_dir, base.Evaluation.SUMMARY_HTML.replace('.html', '.json')
    )
    self.assertTrue(os.path.exists(summary_json))
    summary = pg.load(summary_json, auto_dict=True)
    self.assertIn('Evaluation', summary)
    self.assertEqual(len(summary['Evaluation']), 1)
    self.assertIsNotNone(summary['Evaluation'][0].experiment)
    self.assertIsNotNone(summary['Evaluation'][0].metrics)

  def test_run_wihtout_save(self):
    lm = fake.StaticSequence([
        'Solution(final_answer=2)',
        '3',
    ])
    s = eval_set(
        'run_without_save_test', 'query', schema_fn=answer_schema(), lm=lm)
    s.run(save=False, show_progress=False)

    # Cache will always be saved
    self.assertTrue(
        os.path.exists(os.path.join(s.dir, base.Evaluation.CACHE_JSON)))
    self.assertFalse(
        os.path.exists(os.path.join(s.dir, base.Evaluation.EXPERIMENT_JSON)))
    self.assertFalse(
        os.path.exists(os.path.join(s.dir, base.Evaluation.RESULT_JSON)))
    self.assertFalse(
        os.path.exists(os.path.join(s.dir, base.Evaluation.INDEX_HTML)))
    self.assertFalse(
        os.path.exists(os.path.join(s.dir, base.Evaluation.OOP_FAILURES_HTML)))
    self.assertFalse(
        os.path.exists(
            os.path.join(s.dir, base.Evaluation.NON_OOP_FAILURES_HTML)))

  def test_load(self):
    lm = fake.StaticResponse('Solution(final_answer=2)')
    s = eval_set('loas_test', 'query', schema_fn=answer_schema(), lm=lm)
    s.run(dryrun=True)
    self.assertIsNotNone(s.result)

    s2 = base.load(s.dir)
    self.assertTrue(pg.eq(s, s2))
    self.assertIsNone(s2.result)
    s2.load_result()
    self.assertEqual(s2.result, s.result)

  def test_run_with_filter(self):
    lm = fake.StaticResponse('Solution(final_answer=2)')
    s = eval_set(
        'run_filter_test', pg.oneof(['call', 'query']),
        schema_fn=answer_schema(), lm=lm)
    result = s.run(
        filter=lambda x: x.method == 'query', dryrun=True, summary=False
    )
    self.assertEqual(
        result,
        {
            s.children[0].id: None,
            s.children[1].id: dict(
                experiment_setup=dict(
                    id=s.children[1].id,
                    dir=s.children[1].dir,
                    model='StaticResponse',
                    prompt_template='{{example.question}}',
                    method='query',
                    schema_fn='answer_schema()',
                ),
                cache_stats=dict(
                    use_cache=True, num_queries=2, num_hits=0, num_updates=2
                ),
                metrics=dict(
                    total=2,
                    failures=0,
                    failure_rate=0.0,
                    oop_failures=0,
                    oop_failure_rate=0.0,
                    non_oop_failures=0,
                    non_oop_failure_rate=0.0,
                    failure_breakdown={},
                ),
                usage=s.children[1].result.usage,
            ),
        },
    )

  def test_search_space(self):
    lm = fake.StaticSequence([
        'Solution(final_answer=2)',
        '3',
    ])
    s = base.Evaluation(
        root_dir=tempfile.gettempdir(),
        inputs=base.as_inputs([
            pg.Dict(question='Compute 1 + 1'),
            pg.Dict(question='Compute 1 + 2'),
        ]),
        method='query',
        prompt=pg.oneof([
            lf.Template('{{example.question}}'),
            lf.Template('Hello {{example.question}}'),
        ]),
        schema_fn=answer_schema(),
        lm=lm,
        use_cache=True,
        max_workers=1,
    )
    self.assertEqual(s.children[0].id, f'{s.id}@{s.children[0].hash}')
    self.assertEqual(
        s.children[0].dir, os.path.join(s.root_dir, s.children[0].id)
    )
    # Test persistent hash.
    self.assertEqual(s.hash, 'de23bf31')

    summary = s.run(verbose=True)
    self.assertEqual(len(summary.evaluations), 2)
    self.assertEqual(
        s.result,
        {
            s.children[0].id: dict(
                experiment_setup=dict(
                    id=s.children[0].id,
                    dir=s.children[0].dir,
                    model='StaticSequence',
                    prompt_template='{{example.question}}',
                    method='query',
                    schema_fn='answer_schema()',
                ),
                cache_stats=dict(
                    use_cache=True, num_queries=2, num_hits=0, num_updates=2
                ),
                metrics=dict(
                    total=2,
                    failures=1,
                    failure_rate=0.5,
                    oop_failures=1,
                    oop_failure_rate=0.5,
                    non_oop_failures=0,
                    non_oop_failure_rate=0.0,
                    failure_breakdown={
                        'MappingError.SchemaError.TypeError': 1
                    }
                ),
                usage=s.children[0].result.usage,
            ),
            s.children[1].id: dict(
                experiment_setup=dict(
                    id=s.children[1].id,
                    dir=s.children[1].dir,
                    model='StaticSequence',
                    prompt_template='Hello {{example.question}}',
                    method='query',
                    schema_fn='answer_schema()',
                ),
                cache_stats=dict(
                    use_cache=True, num_queries=2, num_hits=0, num_updates=2
                ),
                metrics=dict(
                    total=2,
                    failures=1,
                    failure_rate=0.5,
                    oop_failures=1,
                    oop_failure_rate=0.5,
                    non_oop_failures=0,
                    non_oop_failure_rate=0.0,
                    failure_breakdown={
                        'MappingError.SchemaError.TypeError': 1
                    }
                ),
                usage=s.children[1].result.usage,
            ),
        },
    )

  def test_call(self):
    lm = fake.StaticSequence(['two'])
    s = eval_set('call_test1', 'call', schema_fn=None, lm=lm)
    self.assertEqual(s.process(s.examples[0]).text, 'two')

    lm = fake.StaticSequence(['two', 'Solution(final_answer=2)'])
    s = eval_set('call_test2', 'call', schema_fn=answer_schema(), lm=lm)
    self.assertEqual(s.process(s.examples[0]).result, Solution(2))

    lm = fake.StaticSequence(['two\n1', 'Solution(final_answer=2)'])

    class CallWithPostProcess(base.Evaluation):
      def call_postprocess(self, lm_response):
        return lm_response.split('\n')[0]

    s = eval_set(
        'call_test3', 'call',
        schema_fn=answer_schema(), lm=lm, cls=CallWithPostProcess,
    )
    self.assertEqual(s.process(s.examples[0]).lm_input.source.text, 'two')

  def test_query(self):
    lm = fake.StaticSequence(['Solution(final_answer=2)'])
    s = eval_set('query_test', 'query', schema_fn=answer_schema(), lm=lm)
    self.assertEqual(s.process(s.examples[0]).result, Solution(2))

    # Test query with fewshot examples.
    lm = fake.StaticSequence(['Solution(final_answer=2)'])
    s = eval_set(
        'basic_test',
        'query',
        schema_fn=answer_schema_with_fewshot_examples(),
        lm=lm,
    )
    m = s.process(s.examples[0])
    self.assertIn('The result of one plus two', m.lm_input.text)

  def test_complete(self):
    lm = fake.StaticSequence(
        ["SolutionForCompletion(question='Compute 1 + 1', final_answer=2)"]
    )
    s = eval_set(
        'complete_test', 'complete', schema_fn=complete_schema(), lm=lm
    )
    self.assertEqual(
        s.process(s.examples[0]).result,
        SolutionForCompletion('Compute 1 + 1', 2)
    )

    # Testing for using a query schema for completion.

    @pg.functor()
    def _answer_schema():

      class Solution2(pg.Object):
        answer: int

      return Solution2

    lm = fake.StaticSequence(
        ["Solution2(question='Compute 1 + 1', answer=2)"],
    )
    s = eval_set(
        'complete_test2', 'complete', schema_fn=_answer_schema(), lm=lm
    )
    self.assertEqual(s.process(s.examples[0]).result.answer, 2)


class SuiteTest(unittest.TestCase):
  """Suite test."""

  def test_run(self):
    lm = fake.StaticSequence([
        'Solution(final_answer=2)',
        '3',
    ] * 5)
    s = base.Suite(
        [
            eval_set('run_test_1', 'query', schema_fn=answer_schema()),
            # A suite of search space. Two of the sub-experiments are identical,
            # thus the result of run_test_2 would include only two keys.
            eval_set('run_test_2',
                     pg.oneof(['call', 'query']),
                     schema_fn=pg.oneof([answer_schema(), answer_schema()])),
        ],
        lm=lm
    )
    # Test for persistent hash.
    self.assertEqual(s.hash, '1c42f93e')
    s.run()
    expected = {
        s.children[0].id: dict(
            experiment_setup=dict(
                id=s.children[0].id,
                dir=s.children[0].dir,
                model='StaticSequence',
                prompt_template='{{example.question}}',
                method='query',
                schema_fn='answer_schema()',
            ),
            cache_stats=dict(
                use_cache=True, num_queries=2, num_hits=0, num_updates=2
            ),
            metrics=dict(
                total=2,
                failures=1,
                failure_rate=0.5,
                oop_failures=1,
                oop_failure_rate=0.5,
                non_oop_failures=0,
                non_oop_failure_rate=0.0,
                failure_breakdown={
                    'MappingError.SchemaError.TypeError': 1
                }
            ),
            usage=s.children[0].result.usage,
        ),
        s.children[1].children[0].id: dict(
            experiment_setup=dict(
                id=s.children[1].children[0].id,
                dir=s.children[1].children[0].dir,
                model='StaticSequence',
                prompt_template='{{example.question}}',
                method='call',
                schema_fn='answer_schema()',
            ),
            cache_stats=dict(
                use_cache=True, num_queries=4, num_hits=0, num_updates=4
            ),
            metrics=dict(
                total=2,
                failures=2,
                failure_rate=1.0,
                oop_failures=2,
                oop_failure_rate=1.0,
                non_oop_failures=0,
                non_oop_failure_rate=0.0,
                failure_breakdown={
                    'MappingError.SchemaError.TypeError': 2
                }
            ),
            usage=s.children[1].children[0].result.usage,
        ),
    }
    self.assertEqual(s.result, expected)


class InputsFrom(unittest.TestCase):
  """Tests for inputs_from."""

  def setUp(self):
    super().setUp()
    pg.symbolic.set_save_handler(pg.symbolic.default_save_handler)
    pg.symbolic.set_load_handler(pg.symbolic.default_load_handler)

  def test_inputs_from_a_single_file(self):
    tmp_dir = tempfile.gettempdir()
    path = os.path.join(tmp_dir, 'input_file.json')
    pg.save([1, 2, 3], path)
    self.assertEqual(base.inputs_from(path)(), [1, 2, 3])

    path = os.path.join(tmp_dir, 'input_file.jsonl')
    with pg.open_jsonl(path, 'w') as f:
      f.add(pg.Dict(x=1))
      f.add(dict(y=2))
    self.assertEqual(
        base.inputs_from(path)(), [pg.Dict(x=1), dict(y=2)]
    )

  def test_inputs_from_multiple_files(self):
    tmp_dir = tempfile.gettempdir()
    path1 = os.path.join(tmp_dir, 'input_file1.json')
    pg.save([1, 2, 3], path1)
    path2 = os.path.join(tmp_dir, 'input_file2.json')
    pg.save([4, 5, 6], path2)
    self.assertEqual(base.inputs_from([path1, path2])(), [1, 2, 3, 4, 5, 6])

  def test_as_inputs(self):
    self.assertEqual(base.as_inputs([1, 2, 3])(), [1, 2, 3])


class TaskA(base.Evaluation):
  pass


class TaskB(base.Evaluation):
  pass


class SummaryTest(unittest.TestCase):

  def _eval_set(self, root_dir):
    return base.Suite(id='select_test', children=[
        TaskA(
            inputs=base.as_inputs([
                pg.Dict(question='Compute 1 + 1'),
            ]),
            method=pg.oneof(['query', 'call']),
            prompt=pg.oneof([
                lf.Template('{{example.question}}'),
                lf.Template('Hello {{example.question}}'),
            ]),
            schema_fn=pg.oneof([
                answer_schema(),
                answer_schema_with_fewshot_examples(),
            ]),
            lm=pg.oneof([
                fake.StaticSequence(['3']),
                fake.StaticResponse('2'),
            ]),
            use_cache=True,
            max_workers=1,
        ),
        TaskB(
            inputs=base.as_inputs([
                pg.Dict(question='Compute 1 + 1'),
            ]),
            method=pg.oneof(['query', 'call']),
            prompt=pg.oneof([
                lf.Template('{{example.question}}'),
            ]),
            schema_fn=pg.oneof([
                answer_schema(),
            ]),
            lm=pg.oneof([
                fake.StaticSequence(['3']),
                fake.StaticResponse('2'),
            ]),
            use_cache=True,
            max_workers=1,
        ),
    ], root_dir=root_dir)

  def test_select(self):
    summary = self._eval_set(None).summary()
    self.assertEqual(len(summary), 2 * 2 * 2 * 2 + 2 * 1 * 1 * 2)

    # Select on task.
    self.assertEqual(len(summary.select(TaskA)), 2 * 2 * 2 * 2)
    self.assertEqual(len(summary.select(TaskB)), 2 * 1 * 1 * 2)

    # Select on LM.
    self.assertEqual(
        len(summary.select(lm=fake.StaticResponse)),
        2 * 2 * 2 * 1 + 2 * 1 * 1 * 1
    )
    self.assertEqual(
        len(summary.select(lm=fake.StaticSequence(['3']))),
        2 * 2 * 2 * 1 + 2 * 1 * 1 * 1
    )
    self.assertEqual(
        len(summary.select(lm=(fake.StaticSequence, fake.StaticResponse))),
        2 * 2 * 2 * 2 + 2 * 1 * 1 * 2
    )

    # Select on method.
    self.assertEqual(
        len(summary.select(method='call')),
        2 * 2 * 2 * 1 + 2 * 1 * 1 * 1
    )
    self.assertEqual(
        len(summary.select(method=('call', 'query'))),
        2 * 2 * 2 * 2 + 2 * 1 * 1 * 2
    )

    # Select on schema.
    self.assertEqual(
        len(summary.select(schema_fn=answer_schema())),
        2 * 2 * 2 * 1 + 2 * 1 * 1 * 2
    )
    self.assertEqual(
        len(summary.select(schema_fn=answer_schema_with_fewshot_examples())),
        2 * 2 * 2 * 1
    )
    self.assertEqual(
        len(summary.select(
            schema_fn=(
                answer_schema(), answer_schema_with_fewshot_examples()))),
        2 * 2 * 2 * 2 + 2 * 1 * 1 * 2
    )

    # Select on completed.
    self.assertEqual(len(summary.select(completed=True)), 0)
    self.assertEqual(
        len(summary.select(completed=False)), 2 * 2 * 2 * 2 + 2 * 1 * 1 * 2)

  def test_from_dirs(self):
    root_dir = os.path.join(tempfile.gettempdir(), 'from_dirs_test')
    s = self._eval_set(root_dir)
    s.run()
    self.assertEqual(
        len(base.Summary.from_dirs(root_dir)), 2 * 2 * 2 * 2 + 2 * 1 * 1 * 2
    )
    self.assertEqual(
        len(base.Summary.from_dirs(root_dir, 'TaskB')), 2 * 1 * 1 * 2
    )
    self.assertEqual(
        len(base.Summary.from_dirs(root_dir, ('TaskA'))), 2 * 2 * 2 * 2
    )

  def test_monitor(self):
    root_dir = os.path.join(tempfile.gettempdir(), 'monitor_test')
    s = self._eval_set(root_dir)
    s.run(summary=False)
    summary_file = os.path.join(root_dir, 'my_summary.html')
    summary = base.monitor(root_dir, summary_file)
    self.assertTrue(all(e.result for e in summary.evaluations))
    self.assertTrue(pg.io.path_exists(summary_file))

  def test_monitor_async(self):
    root_dir = os.path.join(tempfile.gettempdir(), 'monitor_async_test')
    pg.io.mkdirs(root_dir)
    summary_file = os.path.join(root_dir, 'my_summary.html')
    r = base.monitor_async(root_dir, summary_file, expect_new_dirs=True)
    self._eval_set(root_dir).run(summary=False)
    summary = r.stop()
    self.assertTrue(all(e.result for e in summary.evaluations))
    self.assertTrue(pg.io.path_exists(summary_file))


class NamedEvaluationTest(unittest.TestCase):

  def test_named_eval_class(self):

    @base.register('named_eval/class_test')
    class MyEval(base.Evaluation):
      inputs = base.as_inputs([
          pg.Dict(question='Compute 1 + 1'),
      ])
      method = 'query'
      prompt = pg.oneof([
          lf.Template('{{example.question}}'),
          lf.Template('Hello {{example.question}}'),
      ])
      schema_fn = answer_schema()

    [evaluation] = base.get_evaluations('named_eval/class_test')
    self.assertIsInstance(evaluation, MyEval)
    self.assertIsNone(evaluation.dir)
    self.assertIsNone(evaluation.root_dir)
    self.assertIn('named_eval/class_test', base.registered_names())

    with self.assertRaisesRegex(ValueError, 'Unsupported type.*'):
      @base.register('named_eval/bad_class')
      class Foo:  # pylint: disable=unused-variable
        pass

  def test_named_eval_functor(self):

    @base.register('named_eval/functor_test')
    def my_eval():
      return base.Evaluation(
          inputs=base.as_inputs([
              pg.Dict(question='Compute 1 + 1'),
          ]),
          method='query',
          prompt=pg.oneof([
              lf.Template('{{example.question}}'),
              lf.Template('Hello {{example.question}}'),
          ]),
          schema_fn=answer_schema(),
      )

    self.assertTrue(issubclass(my_eval, base.Evaluable))
    [evaluation] = base.get_evaluations('named_eval/functor_test')
    self.assertIn('named_eval/functor_test', base.registered_names())
    self.assertIsInstance(evaluation, my_eval)
    self.assertIsNone(evaluation.root_dir, None)

    self.assertTrue(
        pg.eq(base.get_evaluations('named_eval/functor.*'), [evaluation])
    )
    self.assertEqual(base.get_evaluations('named_eval/non_existent'), [])

    with self.assertRaisesRegex(TypeError, 'The return value .*'):
      @base.register('named_eval/bad_return_type')
      def bad_eval():   # pylint: disable=unused-variable
        return 1

  def test_run(self):
    @base.register('test/run')
    def test_run():  # pylint: disable=unused-variable
      lm = fake.StaticResponse('Solution(final_answer=2)')
      return eval_set('run_test', 'query', schema_fn=answer_schema(), lm=lm)

    e = base.run(
        tempfile.gettempdir(),
        ['test/run'],
        id_regex='run_test.*',
        mode='dryrun',
        print_definition=True,
    )
    self.assertEqual(
        e.leaf_nodes[0].dir,
        os.path.join(tempfile.gettempdir(), e.leaf_nodes[0].id),
    )
    self.assertTrue(
        pg.eq(
            e.leaf_nodes[0].lm, fake.StaticResponse('Solution(final_answer=2)')
        )
    )

    @pg.patcher()
    def bad_lm(unused_eval):  # pylint: disable=unused-variable
      return dict(lm=fake.StaticResponse('efg'))

    e = base.run(
        tempfile.gettempdir(),
        [test_run()],
        filter='Evaluation.*',
        patches=['bad_lm']
    )
    self.assertTrue(pg.eq(e.leaf_nodes[0].lm, fake.StaticResponse('efg')))

    with self.assertRaisesRegex(ValueError, 'No evaluations found'):
      base.run(tempfile.gettempdir(), ['test/non_existent'])


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Image modality."""

import functools
import io
from typing import Any

from langfun.core.modalities import mime

try:
  from PIL import Image as pil_image    # pylint: disable=g-import-not-at-top
  PILImage = pil_image.Image
  pil_open = pil_image.open
except ImportError:
  PILImage = Any

  def pil_open(*unused_args, **unused_kwargs):
    raise RuntimeError(
        'Please install "langfun[mime-pil]" to enable PIL image support.'
    )


class Image(mime.Mime):
  """Image."""

  MIME_PREFIX = 'image'

  @functools.cached_property
  def image_format(self) -> str:
    return self.mime_type.removeprefix(self.MIME_PREFIX + '/')

  def _mime_control_for(self, uri: str) -> str:
    return f'<img src="{uri}">'

  @functools.cached_property
  def size(self) -> tuple[int, int]:
    """Returns the size (width, height) of the image in pixels."""
    img = pil_open(io.BytesIO(self.to_bytes()))
    return img.size

  def to_pil_image(self) -> PILImage:   # pytype: disable=invalid-annotation
    return pil_open(io.BytesIO(self.to_bytes()))

  @classmethod
  def from_pil_image(cls, img: PILImage) -> 'Image':  # pytype: disable=invalid-annotation
    buf = io.BytesIO()
    img.save(buf, format='PNG')
    return cls.from_bytes(buf.getvalue())



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Image drawing for facilitating UI understanding."""

import functools
from typing import Callable
import langfun as lf
from langfun.assistant.capabilities.gui import location
from PIL import Image as pil_image
from PIL import ImageDraw as pil_draw


def blank_image(
    size: tuple[int, int],
    background: tuple[int, int, int] = (0, 0, 0),
    pil: bool = False
    ) -> lf.Image | pil_image.Image:
  """Creates a blank image of given size and background color.

  Args:
    size: The size of the image.
    background: The background color of the image in RGB format.
    pil: If True, the return value will be a `PIL.Image` object, otherwise it
      will be a `lf.Image` object.

  Returns:
    A blank image with requested size and background color.
  """
  image = pil_image.new('RGB', size, background)
  return image if pil else lf.Image.from_pil_image(image)


def draw(
    image: lf.Image | pil_image.Image,
    draw_fn: Callable[[pil_draw.ImageDraw], None],
    ) -> lf.Image | pil_image.Image:
  """Draws on an image with a draw_fn.

  Args:
    image: An `lf.Image` or a `PIL.Image` object.
    draw_fn: A function that takes a `PIL.ImageDraw` object and draws on it.

  Returns:
    The image after drawing. Its type will be the same as the input image.
  """
  is_pil_input = isinstance(image, pil_image.Image)
  if not is_pil_input:
    image = image.to_pil_image()
  draw_fn(pil_draw.Draw(image))
  return image if is_pil_input else lf.Image.from_pil_image(image)


def draw_bboxes(
    image: lf.Image | pil_image.Image,
    bboxes: list[location.BBox],
    line_color: str = 'red',
    line_width: int = 3,
    text: str | None = None,
) -> lf.Image | pil_image.Image:
  """Draws bounding boxes on an image.

  Args:
    image: An `lf.Image` or a `PIL.Image` object.
    bboxes: A list of `location.BBox` objects to draw on the image.
    line_color: The color of the bounding box lines.
    line_width: The width of the bounding box lines.
    text: The text to draw on each bounding box.

  Returns:
    The image after drawing. Its type will be the same as the input image.
  """
  if not bboxes:
    return image

  def _draw_fn(drawing: pil_draw.ImageDraw) -> None:
    for bbox in bboxes:
      drawing.rectangle(
          (bbox.x, bbox.y, bbox.right, bbox.bottom),
          outline=line_color, width=line_width
      )
      if text:
        drawing.text((bbox.x + 5, bbox.y + 5), text, fill=line_color)
  return draw(image, _draw_fn)


def draw_points(
    image: lf.Image | pil_image.Image,
    points: list[location.Coordinate],
    color: str = 'red',
    radius: int = 3,
) -> lf.Image | pil_image.Image:
  """Draws points on an image.

  Args:
    image: An `lf.Image` or a `PIL.Image` object.
    points: A list of `location.Coordinate` objects to draw on the image.
    color: The color of the points.
    radius: The radius of the points.

  Returns:
    The image after drawing. Its type will be the same as the input image.
  """
  if not points:
    return image

  def _draw_fn(drawing: pil_draw.ImageDraw) -> None:
    for point in points:
      drawing.ellipse(
          (
              point.x - radius,
              point.y - radius,
              point.x + radius,
              point.y + radius,
          ),
          fill=color,
      )

  return draw(image, _draw_fn)


def draw_calibration_lines(
    image: lf.Image | pil_image.Image,
    coordinate: location.Coordinate,
    vline_color: str = 'green',
    vline_width: int = 3,
    hline_color: str = 'blue',
    hline_width: int = 3
) -> lf.Image | pil_image.Image:
  """Draws calibration lines for a coordinate in an image.

  Args:
    image: An `lf.Image` or a `PIL.Image` object.
    coordinate: The coordinate to draw calibration lines for.
    vline_color: The color of the vertical line.
    vline_width: The width of the vertical line.
    hline_color: The color of the horizontal line.
    hline_width: The width of the horizontal line.

  Returns:
    The image after drawing. Its type will be the same as the input image.
  """
  def _draw_fn(drawing: pil_draw.ImageDraw) -> None:
    drawing.line(
        (0, coordinate.y, image.size[0], coordinate.y),
        fill=hline_color, width=hline_width
    )
    drawing.line(
        (coordinate.x, 0, coordinate.x, image.size[1]),
        fill=vline_color, width=vline_width
    )
  return draw(image, _draw_fn)


def draw_cursor(
    image: lf.Image | pil_image.Image,
    coordinate: location.Coordinate,
) -> lf.Image | pil_image.Image:
  """Draws a cursor on an image.

  Args:
    image: An `lf.Image` or a `PIL.Image` object.
    coordinate: The coordinate to draw the cursor at.

  Returns:
    The image after drawing. Its type will be the same as the input image.
  """
  is_pil_input = isinstance(image, pil_image.Image)
  if not is_pil_input:
    image = image.to_pil_image()

  image.paste(
      _cursor_image(), (coordinate.x - 3, coordinate.y - 2), _cursor_image()
  )
  return image if is_pil_input else lf.Image.from_pil_image(image)


def draw_ref_points(
    image: lf.Image | pil_image.Image,
    width: int,
    height: int,
    rows: int,
    cols: int,
) -> lf.Image | pil_image.Image:
  """Draw a grid of reference points on an image."""
  delta_x = width // (cols + 1)
  delta_y = height // (rows + 1)
  points = []

  for i in range(rows):
    for j in range(cols):
      x = int((i + 1) * delta_x)
      y = int((j + 1) * delta_y)
      points.append((x, y))

  def _draw_fn(drawing: pil_draw.ImageDraw) -> None:
    for point in points:
      x, y = point
      drawing.ellipse(
          (x - 2, y - 2, x + 2, y + 2), fill='green', outline='green'
      )
      drawing.text((x + 5, y - 6), f'(x={x}, y={y})', fill='green')

  return draw(image, _draw_fn)


def draw_text(
    image: lf.Image | pil_image.Image,
    text: str,
    coordinate: location.Coordinate,
    font_color: str = 'red',
) -> lf.Image | pil_image.Image:
  """Draws text on an image."""
  is_pil_input = isinstance(image, pil_image.Image)
  if not is_pil_input:
    image = image.to_pil_image()

  drawing = pil_draw.ImageDraw(image)
  drawing.text(
      (coordinate.x, coordinate.y),
      text,
      fill=font_color,
      spacing=4,
  )
  return image if is_pil_input else lf.Image.from_pil_image(image)


@functools.cache
def _cursor_image() -> pil_image.Image:
  return lf.Image.from_bytes(
      b"""\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\x19\x00\x00\x00\x1e\x08\x06\x00\x00\x00\xd9\xec\xb5\xdb\x00\x00\x04gIDATx\x9c\xc5\x95_H[g\x18\xc6\x9f\x9c$\'1\xccl\x8ch\xe7\xe2\xb4%\xd2\xea\xa8s\xdd&\xd8Q\x9b\xba\x8d\x8eZA\x10\x0cQ\xca\x84\xad^l\x17e-+\xb8\x1b\x99]\xbb\xe26Ea\xa0\xb2\x8bA\x07\xcat\xa3(Sj\xa7\xad4C+v\x17\xa27\xb6\xb3\x81\xd9\xe9b\xa41mNr\xce\xc9I\x9e]\xa8\x99v\xadZk\xb7\x07^\xbes\xf1\x9d\xe7\xf7\xbd\x7f\xbes\x80\xffR\xe1p\xd8\xa5(J\xf9\xd3\xf0\x16V\xd6\xd9\xd9\xd9lQ\x14\x7f\x88\xc5b_\xf5\xf4\xf4X\x9e\x06\xcc422R\x17\x89D8;;K\x92W\xee\xdc\xb9\xb3{\xbb\xccW2\x01\xc9X(\x14\x82\xd3\xe9DOOO\x91\xddn\xbf"IR\xe9v\x81\x00@\xf4x<5\xc1`\x90V\xab\x95:\x9d\x8e\xb5\xb5\xb5T\x14E%\xf9Y[[\x9bq[ \x83\x83\x83\x9f\x84B!\xee\xd8\xb1\x83\x00\x08\x80\xc5\xc5\xc5\x9c\x9b\x9b#\xc9\xee\x9b7o\xa6?1\xa4\xbf\xbf\xffD8\x1c\xa6\xddnO@\x00p\xd7\xae]\xbcv\xed\x1aIN\x05\x02\x81C[1_\xe9\tUU\x8d\n\x82\x00\x83\xc1\xb0f\x83\xd7\xeb\xc5\xe1\xc3\x87\xd1\xdc\xdc\xbc\xdbj\xb5\xfe\xac\xaa\xea\x89\xadfb\xbcx\xf1\xe2\x07\x8a\xa2\xd0\xe1p\xac\xc9du\x1c;v\x8c\x81@\x80$\xbf\x1b\x1e\x1e~\xfe\xb1!\xed\xed\xedU\xd1h\x94\xd9\xd9\xd9\x8f\x84\x00`^^\x1e\'&&Hr\xcc\xe7\xf3\xbd\xba\x19\xf3\xc4\x08\xab\xaa\xaa\x02\x80\xd1\xb8\xfe \x8d\x8f\x8f\xa3\xb0\xb0\x10\x1d\x1d\x1do\xa4\xa6\xa6\xfe\x12\x0e\x87+7\x0b\xa1\xaa\xaaQ\x00\x14Eq\xc3\x93-..\xa2\xb2\xb2\x12\xa7N\x9d\xb2\x01\xf8\x9e\xe4\x97\r\r\rI\x1b\xbdghnn.\x8d\xc5b\xb1\x82\x82\x82u\xcb\xf5`\x14\x15\x15qff\x86$/y\xbd\xde\x9d\xebe\x02I\x92\xa2\x00h2\x996\xccd\xb5\xae^\xbd\x8a}\xfb\xf6\xa1\xb7\xb7\xf7\xdd\xf4\xf4\xf4\x81\xb9\xb9\xb9\xb7\xfe\x95\xc1\xf2\xcaH$\x12\x15\x04!&\x8a\xa2\xfeQ\x86\x82  ++\x0b\xa2(\xc2b\xb1 ))\tf\xb3\x19\x06\x83\x01\x9d\x9d\x9d\xd8\xbbw\xaf#33\xb37\x14\n}\xd4\xd7\xd7w\xc1\xe5r\xc5VCt\xb2,k\x00\xe2\x1b\xf5\xa4\xa6\xa6\x06UUU\x08\x85B\x7f\x01P\x01h$\xa3\xb2,3\x18\x0cbjj*n6\x9b\xdfKII\x99\x00\xf0\x1b\x00&n^$\x12\xd1\x00\xc4VC,\x16\x0b\xce\x9d;\x87\xee\xeen\x0c\r\r!\x1e\x8f\xe3\xfc\xf9\xf3p\xbb\xdd\x18\x19\x19\xb9TVV\xf6\xa3\xa6iPU5\x8e\xa5\xd2\x1b\x00\xc4\x00\xf8\x00\xfc\t@\x0f@[\xf1\xd3\x1f?~\xbc\x80d\xd0\xedv\x13\x00\xf7\xec\xd9\xc3\xeb\xd7\xaf\x93dtxx\x98F\xa31\xd1\xec3g\xce\x90\xa4\xdf\xe5r\x1d\x04\xf0\xc2r\xa4\x02H\x01`\x03\xf0\x1c\x003\x00\xdd\xea*\xe8].W>\xc9\xbbn\xb7\x9b\xc5\xc5\xc5\xf4\xfb\xfd\x9c\x9e\x9e\xfe\xf5\xf4\xe9\xd3\xf5$#\xd5\xd5\xd5\t\x88\xd5j\xe5\xcc\xcc\x0c%I\xfa\xf6A\xa3\xf5\xa4/))yM\x96e\xff\x8d\x1b7(I\x92:00pA\xaf\xd7\x1f\x05\xf0\xf6\xe8\xe8\xe8O~\xbf\x9f6\x9b-\x01\xaa\xa8\xa8 \xc9\x88\xc7\xe39\xb4i\xc8\x81\x03\x07^\x91e\xf9\xae\xa2(\xf7\xeb\xeb\xeb\xbf\x00P\x00 \x03\x80}\xff\xfe\xfdG5M\xf3555% \x82 \xd0\xe3\xf1\x90\xe4\x95\xb4\xb4\xb4M\xfd\xae\x85#G\x8e\xe4\xde\xbe}{\xb4\xba\xba\xfaS\x00\xafc\xa9\xb6z\x00&\x00;\xbb\xba\xba\xbe\xd64\x8d\xb9\xb9\xb9\tP~~>\xe3\xf18\xa7\xa7\xa7\xdf\xdfT&\x0e\x87\xe3\xa5\x9c\x9c\x9c\x83\x00r\xb0\xd4\xb8\x95\x8b\xaa\x03\xf0lrr\xf2\x9b\x81@`\xb2\xbf\xbf\x9f:\x9d.\x01jmm%\xc9\xdf\xcf\x9e=\x9b\xb6\x11D\x87\xa5iH\xc6C\xa6\x02K\xa3i\xaf\xab\xab\xfb\x98\xa4VZZ\x9a\x80dddP\x92$\xde\xbbw\xef\x1b\xa7\xd3i\xc0\x13\xea\x19\x00\xb9\xb7n\xdd\xba\xec\xf5zy\xf2\xe4Ivvvrrr\x92\x81@@\x9a\x9f\x9f\x9fhiiy\xf9!\x07|,\xe9\x01\xa4\x96\x97\x97W,,,\xfc\xe1\xf3\xf9\xc6\xc7\xc6\xc6:\x1a\x1b\x1b?w:\x9d\x1f\x9aL\xa6w\x00\xbc\xb8\xbco\x8d\x1e\x97j\x06\x90a\xb3\xd9v/,,h\x00\x16\x01\x84\x01H\x00\xee/\x87\x8c\xa5Rn\x19\xb2\xd2;\xcb\xf2\xb3\xb2\x1c\x1a\xfe\xf9\xfc\xff?\xfa\x1b\x9d\xbfI\x0e\x1da[h\x00\x00\x00\x00IEND\xaeB`\x82"""
  ).to_pil_image()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Sampling LM input and responses."""

import functools
from typing import Any, Callable, Generator, Iterator, Tuple, Type, Union

from langfun.core import concurrent
from langfun.core import message as message_lib
from langfun.core.langfunc import LangFunc
import pyglove as pg


def sweep(
    lfun: LangFunc,
    num_examples: int | None = None,
    *,
    max_workers: int = 32,
    silence_on_errors: Union[
        Type[BaseException], Tuple[Type[BaseException], ...], None
    ] = None,
    ignore_examples_with_errors: bool = True,
    **kwargs,
) -> Iterator[
    Tuple[
        message_lib.Message | BaseException,              # LM input.
        Union[message_lib.Message, BaseException, None],  # LM output.
    ],
]:
  """Sweeps the input/output of this LangFunc concurrently.

  Args:
    lfun: An LangFunc object that contains `pg.oneof` as the search space 
      for sampling.
    num_examples: Number of examples to sample.
    max_workers: Max number of concurrent workers to do sampling.
    silence_on_errors: Return None for `lm_input` and `lm_output` when errors
      in this category happen. Otherwise error will be raised during sampling.
    ignore_examples_with_errors: If True, the examples with erratic lm_input
      or lm_output will not be included.
    **kwargs: Keyword arguments as rendering variables.

  Returns:
    An iterator of (lm_input, lm_output).
      `lm_input`/`lm_output` will be None when error happened the error
      matches the type of `silence_on_errors`.
  """
  return _concurrent_sample(
      lfun,
      pg.iter,
      num_examples=num_examples,
      max_workers=max_workers,
      silence_on_errors=silence_on_errors,
      ignore_examples_with_errors=ignore_examples_with_errors,
      **kwargs,
  )


def random_sample(
    lfun: LangFunc,
    num_examples: int | None = None,
    *,
    max_workers: int = 32,
    silence_on_errors: Union[
        Type[BaseException], Tuple[Type[BaseException], ...], None
    ] = None,
    ignore_examples_with_errors: bool = True,
    seed: int | None = None,
    **kwargs,
) -> Iterator[
    Tuple[
        message_lib.Message | BaseException,              # LM input.
        Union[message_lib.Message, BaseException, None],  # LM output.
    ],
]:
  """Random samples the input/output of this LangFunc concurrently.

  Args:
    lfun: An LangFunc object that contains `pg.oneof` as the search space 
      for sampling.
    num_examples: Number of examples to sample.
    max_workers: Max number of concurrent workers to do sampling.
    silence_on_errors: Return None for `lm_input` and `lm_output` when errors
      in this category happen. Otherwise error will be raised during sampling.
    ignore_examples_with_errors: If True, the examples with erratic lm_input
      or lm_output will not be included.
    seed: Random seed.
    **kwargs: Keyword arguments as rendering variables.

  Returns:
    An iterator of (lm_input, lm_output).
      `lm_input`/`lm_output` will be None when error happened the error
      matches the type of `silence_on_errors`.
  """
  return _concurrent_sample(
      lfun,
      functools.partial(pg.random_sample, seed=seed),
      num_examples=num_examples,
      max_workers=max_workers,
      silence_on_errors=silence_on_errors,
      ignore_examples_with_errors=ignore_examples_with_errors,
      **kwargs,
  )


def _concurrent_sample(
    lfun: LangFunc,
    pg_sample_fn: Callable[..., Iterator[Any]],
    num_examples: int | None = None,
    *,
    max_workers: int = 32,
    silence_on_errors: Union[
        Type[BaseException], Tuple[Type[BaseException], ...], None
    ] = None,
    ignore_examples_with_errors: bool = True,
    **kwargs,
) -> Generator[
    Tuple[
        message_lib.Message | BaseException,              # LM input.
        Union[message_lib.Message, BaseException, None],  # LM output.
    ],
    None,
    None,  # Sender type and return type.
]:
  """Concurrently sample the input/output of an LangFunc search space."""

  sampling_space = pg.Dict(
      lfun=lfun,
      kwargs=kwargs,
  )

  if pg.is_deterministic(sampling_space):
    num_examples = num_examples or None
    def repeat_example(example, num_examples=num_examples):
      for _ in range(num_examples):
        yield example.clone()
    pg_sample_fn = repeat_example

  def _error_of(func, *args, **kwargs):
    if silence_on_errors:
      try:
        func(*args, **kwargs)
      except silence_on_errors as e:
        return e
    else:
      func(*args, **kwargs)
    return None

  def _call_fn(example):
    lfun, kwargs = example.lfun, example.kwargs

    error = _error_of(lfun, **kwargs)
    lm_input = lfun.lm_input or error
    lm_output = lfun.lm_output or error
    return lm_input, lm_output

  # Concurrently sample.
  for _, result, error in concurrent.concurrent_map(
      _call_fn,
      pg_sample_fn(sampling_space, num_examples=num_examples),
      silence_on_errors=concurrent.RetryError,
      max_workers=max_workers
  ):
    if error is None:
      lm_input, lm_output = result
    else:
      lm_input, lm_output = error, error
    if (not ignore_examples_with_errors
        or not (isinstance(lm_input, BaseException)
                or isinstance(lm_output, BaseException))):
      yield lm_input, lm_output



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interface for memories."""

import abc
from typing import Any

from langfun.core import template as template_lib
from langfun.core.component import Component
from langfun.core.natural_language import NaturalLanguageFormattable


class Memory(NaturalLanguageFormattable, Component):
  """Interface for memory.

  The role of memory is to form the conversation context within a prompt.
  Instead of simply dumping all the historical messages to the context, the
  abstraction of memory allows agent developers to select relevant pieces of
  information, summary such information to best direct LLMs' attentions and
  utilize their context window.
  """

  def natural_language_format(self) -> str:
    return self.recollect().render().text

  def recollect(self, **kwargs) -> template_lib.Template:
    """Recollects a message from the memory."""
    return template_lib.Template.from_value(self._recollect(**kwargs))

  def remember(
      self,
      value: Any,
      **kwargs
  ) -> None:
    self._remember(value, **kwargs)

  def reset(self, **kwargs) -> None:
    """Resets the memory."""
    self._reset(**kwargs)

  @abc.abstractmethod
  def _recollect(self, **kwargs) -> str | template_lib.Template:
    """Recollects a message from the memory."""

  @abc.abstractmethod
  def _remember(self, value: Any, **kwargs) -> None:
    """Remembers the input message and the output message."""

  @abc.abstractmethod
  def _reset(self, **kwargs) -> None:
    """Reset the memory."""



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""LLM-based function generation."""

import functools
import inspect
import re
from typing import Any, Callable, Literal, Optional, Tuple

from langfun.core import language_model
from langfun.core import template
from langfun.core.coding import python
from langfun.core.structured import querying
import pyglove as pg


def unittest_gen(signature, lm, num_retries=1):
  """Generates unit tests for a python function signature."""

  class UnitTest(pg.Object):
    """A valid unit test for a python function."""

    input: dict[str, Any]
    expected_output: Any

  class PythonFunctionSignature(pg.Object):
    signature: str

  unittest_examples = None
  for _ in range(num_retries):
    r = querying.query(
        PythonFunctionSignature(signature=signature),
        list[UnitTest],
        lm=lm,
        default=None,
    )
    if isinstance(r, list) and r:
      unittest_examples = []
      for unit_test in r:
        unittest_examples.append((unit_test.input, unit_test.expected_output))
      break

  return unittest_examples


def unittest_with_test_cases(f, unittests):
  """Applies unit tests to a python function to be tested."""
  if not unittests:
    raise ValueError(f"No unit tests provided: {unittests}")

  for unit_test in unittests:
    inputs = unit_test[0]
    if isinstance(inputs, dict):
      actual = f(**inputs)
    elif isinstance(inputs, tuple):
      actual = f(*inputs)
    else:
      actual = f(inputs)

    expected = unit_test[1]
    assert (
        actual == expected
    ), f"Test FAILED: Inputs: {inputs}, Expected: {expected}, Actual: {actual}"


def _function_gen(
    func: Callable[..., Any],
    context: dict[str, Any],
    signature: str,
    lm: language_model.LanguageModel,
    num_retries: int = 1,
    unittest: Optional[
        Callable[[Callable[..., Any]], None]
        | list[Tuple[Any, Any]]
        | Literal["auto"]
    ] = None,
    unittest_num_retries: int = 1,
):
  """Generates a python function with LLM and verify its quality with unit testing."""

  class PythonFunctionPrompt(template.Template):
    r"""A template for a python function generation.

    Please reply to the last PYTHON_FUNCTION_SIGNATURE with a self-sufficient,
    error-free, and efficiently coded PYTHON_FUNCTION, crafted to the standards
    of a world-class programmer.

      PYTHON_FUNCTION_SIGNATURE:
        ```python
        def calculate_area_circle(radius: float) -> float:
        \"\"\"Calculates the area of a circle given its radius.

        Args:
            radius: The radius of the circle.

        Returns:
            The area of the circle.
        \"\"\"
        ```

      PYTHON_FUNCTION:
        ```python
        def calculate_area_circle(radius: float) -> float:
        \"\"\"Calculates the area of a circle given its radius.

        Args:
            radius: The radius of the circle.

        Returns:
            The area of the circle.
        \"\"\"
        import math

        area = math.pi * radius**2
        return area
        ```

      PYTHON_FUNCTION_SIGNATURE:
        ```python
        {{signature}}
        ```

      PYTHON_FUNCTION:
    """

  unittest_examples = None
  if unittest == "auto":
    unittest_examples = unittest_gen(
        signature, lm=lm, num_retries=unittest_num_retries
    )
  elif isinstance(unittest, list):
    unittest_examples = unittest

  last_error = None
  for _ in range(num_retries):
    try:
      source_code = querying.query(
          PythonFunctionPrompt(signature=signature), lm=lm
      )
      f = python.evaluate(source_code, global_vars=context)

      # Check whether the sigantures are the same.
      if inspect.signature(f) != inspect.signature(func):
        raise python.CodeError(
            code=source_code,
            cause=TypeError(
                f"Signature mismatch: Expected: {inspect.signature(func)}, "
                f"Actual: {inspect.signature(f)}.",
            ),
        )

      if callable(unittest):
        unittest(f)
      elif unittest_examples:
        unittest_with_test_cases(f, unittest_examples)

      return f, source_code
    except python.CodeError as e:
      last_error = e
      pg.logging.warning(
          f"Bad code generated: {e}",
      )
  raise last_error


def _process_signature(signature):
  # Remove the decorator.
  pattern = r"^\@.*function_gen.*$"
  signature = re.sub(pattern, "", signature, flags=re.MULTILINE)
  # Remove the possible 'pass' in an empty function.
  pattern = r"^\s*pass\s*$"
  signature = re.sub(pattern, "", signature, flags=re.MULTILINE)
  return signature.strip()


def function_gen(
    lm: language_model.LanguageModel,
    cache_filename: str | None = None,
    num_retries: int = 1,
    unittest: Optional[
        Callable[[Callable[..., Any]], None]
        | list[Tuple[Any, Any]]
        | Literal["auto"]
    ] = None,
    unittest_num_retries: int = 1,
):
  """A decorator for automating function generation using a language model.

  This decorator should be applied to functions that are not yet implemented. It
  facilitates the implementation via the specified LLM, ensuring
  quality through unit tests.

  Args:
      lm (lf.LanguageModel): The language model used for generating function
        implementations.
      cache_filename (str | None): Optional. The path of the file where
        generated function implementations are loaded from or saved to.
      num_retries (int): Maximum number of attempts the language model should
        make to generate a suitable function implementation.
      unittest: This optional parameter enables the definition of custom unit
        tests. You can either provide a list of test cases as tuples of inputs
        and outputs, or a function that throws an error if a test fails, or let
        LLM automatically create the unit test cases. If a generated function is
        and returned, it should pass all the unittests.
      unittest_num_retries: If unittest is set to "auto", this parameter
        specifies the number of times the LLM's attempts to generate unit test
        cases.

  Returns:
      The implemented function object.
  """

  def _decorate(func):
    setattr(func, "__function__", None)
    setattr(func, "__source_code__", None)

    # Prepare the globals/locals for the generated code to be evaluated against.
    callstack = inspect.stack()
    assert len(callstack) > 1
    context = dict(callstack[1][0].f_globals)
    context.update(callstack[1][0].f_locals)
    context.pop(func.__name__, None)

    @functools.wraps(func)
    def lm_generated_func(*args, **kwargs):
      if func.__function__ is not None:
        return func.__function__(*args, **kwargs)

      signature = _process_signature(inspect.getsource(func))
      cache = pg.Dict()
      if cache_filename is not None:
        try:
          cache = pg.load(cache_filename)
        except FileNotFoundError:
          pg.logging.warning(
              "Creating a new cache as cache file '%s' does not exist.",
              cache_filename,
          )

        if signature in cache:
          func.__source_code__ = cache[signature]
          func.__function__ = python.evaluate(
              func.__source_code__, global_vars=context
          )
          return func.__function__(*args, **kwargs)

      func.__function__, func.__source_code__ = _function_gen(
          func,
          context,
          signature,
          lm,
          num_retries=num_retries,
          unittest=unittest,
          unittest_num_retries=unittest_num_retries,
      )
      if cache_filename is not None:
        cache[signature] = func.__source_code__
        cache.save(cache_filename)
      return func.__function__(*args, **kwargs)

    lm_generated_func.__name__ = func.__name__
    lm_generated_func.__qualname__ = func.__qualname__
    lm_generated_func.__module__ = func.__module__
    lm_generated_func.source = lambda: func.__source_code__
    return lm_generated_func

  return _decorate



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Self-play."""

import abc
from typing import Annotated, Union
import langfun.core as lf
from langfun.core.templates.conversation import Conversation


@lf.use_init_args(['preamble'])
class SelfPlay(Conversation):
  """LM-based self play."""

  max_turns: Annotated[
      int | None,
      (
          'Max number of turns to play. If None, the self play will continue '
          'until StopIteration is raised from the `next_turn` method.'
      ),
  ] = None

  @property
  def num_turns(self) -> int:
    """Returns number of turns currently played."""
    return self._num_turns

  def __call__(self, **kwargs) -> lf.Message:
    """Starts the play."""
    while self.max_turns is None or self._num_turns < self.max_turns:
      output = self.step(**kwargs)
      if output is None:
        break
    return self._last_response

  def step(self, **kwargs) -> lf.Message | None:
    """Play the next step and return the response."""
    with lf.context(**kwargs):
      input_message = self.next_turn(self._last_response)
      if input_message is None:
        return None
      self._last_response = super().__call__(input_message=input_message)
    self._num_turns += 1
    return self._last_response

  def reset(self):
    super().reset()
    self._num_turns = 0
    self._last_response = None

  @abc.abstractmethod
  def next_turn(
      self, last_response: lf.Message | None) -> Union[None, str, lf.Message]:
    """Returns the `input_message` for the next turn.

    Args:
      last_response: Last LM response. If None, then this will be the first
        message to send.

    Returns:
      The `input_message` for the next turn. None if there is no next turn.
    """



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Messages that are exchanged between users and agents."""

import abc
import collections
from collections.abc import Mapping
import contextlib
import functools
import inspect
import io
from typing import Annotated, Any, ClassVar, Optional, Type, Union

from langfun.core import modality
from langfun.core import natural_language
import pyglove as pg


class Message(
    natural_language.NaturalLanguageFormattable,
    pg.Object,
    pg.views.HtmlTreeView.Extension
):
  """Message.

  ``Message`` is the protocol for users and the system to interact with
  LLMs. It consists of a text in the form of natural language, 
  an identifier of the sender, and a dictionary of Python values as structured
  meta-data.

  The subclasses of ``Message`` represent messages sent from different roles.
  Agents may use the roles to decide the orchastration logic.
  """

  #
  # Constants.
  #

  # PyGlove flag to allow modifications to the member attributes through
  # assignment.
  allow_symbolic_assignment = True

  # Constant that refers to the special key used to access the message itself
  # through `Message.get`.
  PATH_ROOT = ''

  # Constant that refers to the special key used to access the message text
  # through `Message.get`.
  PATH_TEXT = 'text'

  # Constant that refers to the special key for describing the structured
  # information extracted as the result of the response.
  PATH_RESULT = 'result'

  # Constant that tags an LM input.
  TAG_LM_INPUT = 'lm-input'

  # Constant that tags an LM response (before output_transform).
  TAG_LM_RESPONSE = 'lm-response'

  # Constant that tags an LM output message (after output_transform).
  TAG_LM_OUTPUT = 'lm-output'

  # Constant that tags a message that is generated by `LangFunc.render`
  TAG_RENDERED = 'rendered'

  # Constant that tags an transformed message.
  TAG_TRANSFORMED = 'transformed'

  #
  # Members
  #

  text: Annotated[str, 'The natural language representation of the message.']

  sender: Annotated[str, 'The sender of the message.']

  metadata: Annotated[
      dict[str, Any],
      (
          'The metadata associated with the message, '
          'which chould carry structured data, such as tool function input. '
          'It is a `pg.Dict` object whose keys can be accessed by attributes.'
      ),
  ] = pg.Dict()

  tags: Annotated[
      list[str],
      (
          'A list of tags associated with the message. '
          'Tags are useful for filtering messages along the source chain. '
      )
  ] = []

  # NOTE(daiyip): Explicit override __init__ for allowing metadata via **kwargs.
  @pg.explicit_method_override
  def __init__(
      self,
      text: str,
      *,
      # Default sender is specified in subclasses.
      sender: str | pg.object_utils.MissingValue = pg.MISSING_VALUE,
      metadata: dict[str, Any] | None = None,
      tags: list[str] | None = None,
      source: Optional['Message'] = None,
      # The rest are `pg.Object.__init__` arguments.
      allow_partial: bool = False,
      sealed: bool = False,
      root_path: pg.KeyPath | None = None,
      **kwargs
  ) -> None:
    """Constructor.

    Args:
      text: The text in the message.
      sender: The sender name of the message.
      metadata: Structured meta-data associated with this message.
      tags: Tags for the message.
      source: The source message of the current message.
      allow_partial: If True, the object can be partial.
      sealed: If True, seal the object from future modification (unless under a
        `pg.seal(False)` context manager). If False, treat the object as
        unsealed. If None, it's determined by `cls.allow_symbolic_mutation`.
      root_path: The symbolic path for current object. By default it's None,
        which indicates that newly constructed object does not have a parent.
      **kwargs: key/value pairs that will be inserted into metadata.
    """
    metadata = metadata or {}
    metadata.update(kwargs)
    super().__init__(
        text=text,
        metadata=metadata,
        tags=tags or [],
        sender=sender,
        allow_partial=allow_partial,
        sealed=sealed,
        root_path=root_path,
    )
    self._source = source

  @classmethod
  def from_value(
      cls,
      value: Union[str, 'Message', Any],
      *,
      format: str | None = None,    # pylint: disable=redefined-builtin
      **kwargs
  ) -> 'Message':
    """Creates a message from a str, message or object of registered format.

    Example:
      # Create a user message from a str.
      lf.UserMessage.from_value('hi')

      # Create a user message from a multi-modal object.
      lf.UserMessage.from_value(lf.Image.from_uri('...'))

      # Create a user message from OpenAI API format.
      lf.Message.from_value(
          {
              'role': 'user',
              'content': [{'type': 'text', 'text': 'hi'}],
          },
          format='openai.api',
      )

    Args:
      value: The value to create a message from.
      format: The format ID to convert to. If None, the conversion will be
        performed according to the type of `value`. Otherwise, the converter
        registered for the format will be used.
      **kwargs: The keyword arguments passed to the __init__ of the converter.

    Returns:
      A message created from the value.
    """
    if isinstance(value, modality.Modality):
      return cls('<<[[object]]>>', object=value)
    if isinstance(value, Message):
      return value
    if isinstance(value, str):
      return cls(value)
    value_type = type(value)
    if format is None:
      converter = MessageConverter.get_by_type(value_type, **kwargs)
    else:
      converter = MessageConverter.get_by_format(format, **kwargs)
      if (converter.OUTPUT_TYPE is not None
          and not isinstance(value, converter.OUTPUT_TYPE)):
        raise ValueError(f'{format!r} is not applicable to {value!r}.')
    return converter.from_value(value)

  #
  # Conversion to other formats or types.
  #

  def as_format(self, format_or_type: str | type[Any], **kwargs) -> Any:
    """Converts the message to a registered format or type.

    Example:

      m = lf.Template('What is this {{image}}?').render()
      m.as_format('openai')  # Convert to OpenAI message format.
      m.as_format('gemini')  # Convert to Gemini message format.
      m.as_format('anthropic')  # Convert to Anthropic message format.

    Args:
      format_or_type: The format ID or type to convert to.
      **kwargs: The conversion arguments.

    Returns:
      The converted object according to the format or type.
    """
    return MessageConverter.get(format_or_type, **kwargs).to_value(self)

  @classmethod
  def convertible_formats(cls) -> list[str]:
    """Returns supported format for message conversion."""
    return MessageConverter.convertible_formats()

  @classmethod
  def convertible_types(cls) -> list[str]:
    """Returns supported types for message conversion."""
    return MessageConverter.convertible_types()

  #
  # Unified interface for accessing text, result and metadata.
  #

  def set(self, key_path: str | pg.KeyPath, value: Any) -> None:
    """Sets a value by key path.

    This the unified interface for set values in Message.

    Examples::

      m = lf.AIMessage('foo', metadata=dict(x=dict(k=[0, 1]), y=2))
      m.set('text', 'bar')
      m.set('x.k[0]', 1)
      m.set('y', pg.MISSING_VALUE)   # Delete.
      assert m == lf.AIMessage('bar', metadata=dict(x=dict(k=[1, 1])))

    Args:
      key_path: A string or a ``pg.KeyPath`` object for locating the value to
        update. For example: `a.b`, `x[0].a`, `text` is a special key that sets
        the text of the message.
      value: The new value for the location.
    """
    if key_path == Message.PATH_TEXT:
      self.rebind({key_path: value}, raise_on_no_change=False)
    else:
      self.metadata.rebind({key_path: value}, raise_on_no_change=False)

  def get(self, key_path: str | pg.KeyPath, default: Any = None) -> Any:
    """Gets text or metadata by key path.

    This is the unified interface to query text or metadata values.

    Args:
      key_path: A string like 'a.x', 'b[0].y' to access metadata value in
        hierarchy. 'text' is a special key that returns the text of the message.
      default: The default value if the key path is not found in `self`.

    Returns:
      The value for the path if found, otherwise the default value.
    """
    if not key_path:
      return self
    if key_path == Message.PATH_TEXT:
      return self.text
    else:
      v = self.metadata.sym_get(key_path, default, use_inferred=True)
      return v.value if isinstance(v, pg.Ref) else v

  #
  # API for accessing the structured result and error.
  # `result` represents the structured output the message - like the return
  # value of a regular Python function. It is stored in metadata with a special
  # key `result`. By default it's None, and should be produced by transforms
  # upon the message text.

  @property
  def result(self) -> Any:
    """Gets the structured result of the message."""
    return self.get(Message.PATH_RESULT, None)

  @result.setter
  def result(self, value: Any) -> None:
    """Sets the structured result of the message."""
    self.set(Message.PATH_RESULT, value)

  #
  # Update and error tracking.
  #

  def _on_init(self):
    super()._on_init()
    self._updates = {}
    self._errors = []

  def _on_change(self, field_updates: dict[pg.KeyPath, pg.FieldUpdate]) -> None:
    super()._on_change(field_updates)
    self._updates.update(field_updates)

  @property
  def modified(self) -> bool:
    """Returns True if the message has been modified in current update scope."""
    return bool(self._updates)

  @property
  def updates(self) -> dict[pg.KeyPath, pg.FieldUpdate]:
    """Returns the updates of the message in current update scope."""
    return self._updates

  @property
  def has_errors(self) -> bool:
    """Returns True if there is an error in current update scope."""
    return bool(self._errors)

  @property
  def errors(self) -> list[Any]:
    """Returns the errors of the message in current update scope."""
    return self._errors

  @contextlib.contextmanager
  def update_scope(self):
    """Context manager to create a update scope."""
    accumulated_updates = self._updates
    accumulated_errors = self._errors

    try:
      self._updates, self._errors = {}, []
      yield
    finally:
      accumulated_updates.update(self._updates)
      self._updates = accumulated_updates

      accumulated_errors.extend(self._errors)
      self._errors = accumulated_errors

  def apply_updates(self, updates: dict[pg.KeyPath, pg.FieldUpdate]) -> None:
    """Updates this message with delta."""
    delta = {k: v.new_value for k, v in updates.items()}

    # Rebind will trigger _on_change, which inserts the updates
    # to current message' updates.
    self.rebind(delta, raise_on_no_change=False)

  #
  # API for supporting modalities.
  #

  @property
  def text_with_modality_hash(self) -> str:
    """Returns text with modality object placeheld by their 8-byte MD5 hash."""
    parts = [self.text]
    for name, modality_obj in self.referred_modalities().items():
      parts.append(
          f'<{name}>{modality_obj.hash}</{name}>'
      )
    return ''.join(parts)

  def get_modality(
      self, var_name: str, default: Any = None, from_message_chain: bool = True
  ) -> modality.Modality | None:
    """Gets the modality object referred in the message.

    Args:
      var_name: The referred variable name for the modality object.
      default: default value.
      from_message_chain: If True, the look up will be performed from the
        message chain. Otherwise it will be performed in current message.

    Returns:
      A modality object if found, otherwise None.
    """
    obj = self.get(var_name, None)
    if isinstance(obj, modality.Modality):
      return obj
    elif obj is None and self.source is not None:
      return self.source.get_modality(var_name, default, from_message_chain)
    return default

  def referred_modalities(self) -> dict[str, modality.Modality]:
    """Returns modality objects attached on this message."""
    chunks = self.chunk()
    return {
        m.referred_name: m for m in chunks if isinstance(m, modality.Modality)
    }

  def chunk(self, text: str | None = None) -> list[str | modality.Modality]:
    """Chunk a message into a list of str or modality objects."""
    chunks = []

    def add_text_chunk(text_piece: str) -> None:
      if text_piece:
        chunks.append(text_piece)
    if text is None:
      text = self.text

    chunk_start = 0
    ref_end = 0
    while chunk_start < len(text):
      ref_start = text.find(modality.Modality.REF_START, ref_end)
      if ref_start == -1:
        add_text_chunk(text[chunk_start:].strip(' '))
        break

      var_start = ref_start + len(modality.Modality.REF_START)
      ref_end = text.find(modality.Modality.REF_END, var_start)
      if ref_end == -1:
        add_text_chunk(text[chunk_start:])
        break

      var_name = text[var_start:ref_end].strip()
      var_value = self.get_modality(var_name)
      if var_value is not None:
        add_text_chunk(text[chunk_start:ref_start].strip(' '))
        chunks.append(var_value)
        chunk_start = ref_end + len(modality.Modality.REF_END)
    return chunks

  @classmethod
  def from_chunks(
      cls, chunks: list[str | modality.Modality], separator: str = ' '
  ) -> 'Message':
    """Assembly a message from a list of string or modality objects."""
    fused_text = io.StringIO()
    ref_index = 0
    metadata = dict()
    last_char = None
    for i, chunk in enumerate(chunks):
      if i > 0 and last_char not in ('\t', ' ', '\n', None):
        fused_text.write(separator)
      if isinstance(chunk, str):
        fused_text.write(chunk)
        if chunk:
          last_char = chunk[-1]
        else:
          last_char = None
      else:
        assert isinstance(chunk, modality.Modality), chunk
        var_name = f'obj{ref_index}'
        fused_text.write(modality.Modality.text_marker(var_name))
        last_char = modality.Modality.REF_END[-1]
        # Make a reference if the chunk is already owned by another object
        # to avoid copy.
        metadata[var_name] = pg.maybe_ref(chunk)
        ref_index += 1
    return cls(fused_text.getvalue().strip(), metadata=metadata)

  #
  # Tagging
  #

  def tag(self, tag: str) -> None:
    if tag not in self.tags:
      with pg.notify_on_change(False):
        self.tags.append(tag)

  def has_tag(self, tag: str | tuple[str, ...]) -> bool:
    if isinstance(tag, str):
      return tag in self.tags
    return any(t in self.tags for t in tag)

  #
  # Message source chain.
  #

  @property
  def source(self) -> Optional['Message']:
    """Returns the source message."""
    return self._source

  @source.setter
  def source(self, source: 'Message') -> None:
    """Sets the source message."""
    self._source = source

  @property
  def root(self) -> 'Message':
    """Returns the root of this message."""
    root = self
    while root.source is not None:
      root = root.source
    return root

  def trace(self, tag: str | None = None) -> list['Message']:
    """Returns the chain of source messages filtered by tag."""
    message_chain = []
    current = self

    while current is not None:
      if tag is None or tag in current.tags:
        message_chain.append(current)
      current = current.source
    return list(reversed(message_chain))

  @property
  def lm_responses(self) -> list['Message']:
    """Returns a chain of LM responses starting from the first LM call."""
    return self.trace(Message.TAG_LM_RESPONSE)

  @property
  def lm_inputs(self) -> list['Message']:
    """Returns a chain of LM inputs starting from the first LM call."""
    return self.trace(Message.TAG_LM_INPUT)

  @property
  def lm_outputs(self) -> list['Message']:
    """Returns a chain of LM inputs starting from the first LM call."""
    return self.trace(Message.TAG_LM_OUTPUT)

  def last(self, tag: str) -> Optional['Message']:
    """Return the last message wih certain tag."""
    current = self
    while current is not None:
      if tag in current.tags:
        return current
      current = current.source
    return None

  @property
  def lm_response(self) -> Optional['Message']:
    """Returns the latest LM raw response."""
    return self.last(Message.TAG_LM_RESPONSE)

  @property
  def lm_input(self) -> Optional['Message']:
    """Returns the latest LM input."""
    return self.last(Message.TAG_LM_INPUT)

  @property
  def lm_output(self) -> Optional['Message']:
    """Returns the latest LM output."""
    return self.last(Message.TAG_LM_OUTPUT)

  #
  # Other methods.
  #

  def natural_language_format(self) -> str:
    return self.text

  def __eq__(self, other: Any) -> bool:
    if isinstance(other, str):
      return self.text == other
    if isinstance(other, self.__class__):
      return (self.text == other.text
              and self.sender == other.sender
              and self.metadata == other.metadata)
    return False

  def __hash__(self) -> int:
    return hash(self.text)

  def __getattr__(self, key: str) -> Any:
    if key not in self.metadata:
      raise AttributeError(key)
    v = self.metadata[key]
    return v.value if isinstance(v, pg.Ref) else v

  def _html_tree_view_content(
      self,
      *,
      view: pg.views.HtmlTreeView,
      root_path: pg.KeyPath | None = None,
      collapse_level: int | None = None,
      extra_flags: dict[str, Any] | None = None,
      **kwargs,
  ) -> pg.Html:
    """Returns the HTML representation of the message.

    Args:
      view: The HTML tree view.
      root_path: The root path of the message.
      collapse_level: The global collapse level.
      extra_flags: Extra flags to control the rendering.
        - source_tag: tags to filter source messages. If None, the entire
          source chain will be included.
        - include_message_metadata: Whether to include the metadata of the
          message.
        - collapse_modalities_in_text: Whether to collapse the modalities in the
          message text.
        - collapse_llm_usage: Whether to collapse the usage in the message.
        - collapse_message_result_level: The level to collapse the result in the
          message.
        - collapse_message_metadata_level: The level to collapse the metadata in
          the message.
        - collapse_source_message_level: The level to collapse the source in the
          message.
        - collapse_level: The global collapse level.
      **kwargs: Omitted keyword arguments.

    Returns:
      The HTML representation of the message content.
    """
    extra_flags = extra_flags if extra_flags is not None else {}

    include_message_metadata: bool = extra_flags.get(
        'include_message_metadata', True
    )
    source_tag: str | tuple[str, ...] | None = extra_flags.get(
        'source_tag', ('lm-input', 'lm-output')
    )
    collapse_modalities_in_text: bool = extra_flags.get(
        'collapse_modalities_in_text', True
    )
    collapse_llm_usage: bool = extra_flags.get(
        'collapse_llm_usage', False
    )
    collapse_message_result_level: int | None = extra_flags.get(
        'collapse_message_result_level', 1
    )
    collapse_message_metadata_level: int | None = extra_flags.get(
        'collapse_message_metadata_level', 1
    )
    collapse_source_message_level: int | None = extra_flags.get(
        'collapse_source_message_level', 1
    )
    passthrough_kwargs = view.get_passthrough_kwargs(**kwargs)
    def render_tags():
      return pg.Html.element(
          'div',
          [pg.Html.element('span', [tag]) for tag in self.tags],
          css_classes=['message-tags'],
      )

    def render_message_text():
      maybe_reformatted = self.get('formatted_text')
      referred_chunks = {}
      s = pg.Html('<div class="message-text">')
      for chunk in self.chunk(maybe_reformatted):
        if isinstance(chunk, str):
          s.write(s.escape(chunk))
        else:
          assert isinstance(chunk, modality.Modality), chunk
          child_path = pg.KeyPath(['metadata', chunk.referred_name], root_path)
          s.write(
              pg.Html.element(
                  'div',
                  [
                      view.render(
                          chunk,
                          name=chunk.referred_name,
                          root_path=child_path,
                          collapse_level=(
                              0 if collapse_modalities_in_text else 1
                          ),
                          extra_flags=dict(
                              display_modality_when_hover=True,
                          ),
                          **passthrough_kwargs,
                      )
                  ],
                  css_classes=['modality-in-text'],
              )
          )
          referred_chunks[chunk.referred_name] = chunk
      s.write('</div>')
      return s

    def render_result():
      if 'result' not in self.metadata:
        return None
      child_path = pg.KeyPath(['metadata', 'result'], root_path)
      return pg.Html.element(
          'div',
          [
              view.render(
                  self.result,
                  name='result',
                  root_path=child_path,
                  collapse_level=view.get_collapse_level(
                      (collapse_level, -1),
                      collapse_message_result_level,
                  ),
                  extra_flags=extra_flags,
                  **passthrough_kwargs,
              )
          ],
          css_classes=['message-result'],
      )

    def render_usage():
      if 'usage' not in self.metadata:
        return None
      child_path = pg.KeyPath(['metadata', 'usage'], root_path)
      return pg.Html.element(
          'div',
          [
              view.render(
                  self.usage,
                  name='llm usage',
                  key_style='label',
                  root_path=child_path,
                  collapse_level=view.get_collapse_level(
                      (collapse_level, -1),
                      0 if collapse_llm_usage else 1,
                  ),
                  extra_flags=extra_flags,
                  **view.get_passthrough_kwargs(
                      remove=['key_style'], **kwargs
                  ),
              )
          ],
          css_classes=['message-usage'],
      )

    def render_source_message():
      source = self.source
      while (source is not None
             and source_tag is not None
             and not source.has_tag(source_tag)):
        source = source.source
      if source is not None:
        child_path = pg.KeyPath('source', root_path)
        child_extra_flags = extra_flags.copy()
        child_extra_flags['collapse_source_message_level'] = (
            view.get_collapse_level(
                (collapse_source_message_level, -1), 0,
            )
        )
        return view.render(
            self.source,
            name='source',
            root_path=child_path,
            collapse_level=view.get_collapse_level(
                (collapse_level, -1),
                collapse_source_message_level,
            ),
            extra_flags=child_extra_flags,
            **passthrough_kwargs,
        )
      return None

    def render_metadata():
      if not include_message_metadata:
        return None
      child_path = pg.KeyPath('metadata', root_path)
      return pg.Html.element(
          'div',
          [
              view.render(
                  self.metadata,
                  css_classes=['message-metadata'],
                  exclude_keys=['usage', 'result'],
                  name='metadata',
                  root_path=child_path,
                  collapse_level=view.get_collapse_level(
                      (collapse_level, -1),
                      collapse_message_metadata_level,
                  ),
                  **view.get_passthrough_kwargs(
                      remove=['exclude_keys'], **kwargs
                  ),
              )
          ],
          css_classes=['message-metadata'],
      )

    return pg.Html.element(
        'div',
        [
            render_tags(),
            render_message_text(),
            render_result(),
            render_usage(),
            render_metadata(),
            render_source_message(),
        ],
        css_classes=['complex_value'],
    )

  @classmethod
  @functools.cache
  def _html_tree_view_config(cls) -> dict[str, Any]:
    return pg.views.HtmlTreeView.get_kwargs(
        super()._html_tree_view_config(),
        dict(
            css_classes=['lf-message'],
        )
    )

  @classmethod
  @functools.cache
  def _html_tree_view_css_styles(cls) -> list[str]:
    return super()._html_tree_view_css_styles() + [
        """
        /* Langfun Message styles.*/
        [class^="message-"] > details {
            margin: 0px 0px 5px 0px;
            border: 1px solid #EEE;
        }
        .lf-message.summary-title::after {
            content: ' 💬';
        }
        details.pyglove.ai-message {
            border: 1px solid blue;
            color: blue;
        }
        details.pyglove.user-message {
            border: 1px solid green;
            color: green;
        }
        .message-tags {
            margin: 5px 0px 5px 0px;
            font-size: .8em;
        }
        .message-tags > span {
            border-radius: 5px;
            background-color: #CCC;
            padding: 3px;
            margin: 0px 2px 0px 2px;
            color: white;
        }
        .message-text {
            padding: 20px;
            margin: 10px 5px 10px 5px;
            font-style: italic;
            white-space: pre-wrap;
            border: 1px solid #EEE;
            border-radius: 5px;
            background-color: #EEE;
        }
        .modality-in-text {
            display: inline-block;
        }
        .modality-in-text > details.pyglove {
            display: inline-block;
            font-size: 0.8em;
            border: 0;
            background-color: #A6F1A6;
            margin: 0px 5px 0px 5px;
        }
        .message-result {
            color: dodgerblue;
        }
        .message-usage {
            color: orange;
        }
        .message-usage .object-key.str {
            border: 1px solid orange;
            background-color: orange;
            color: white;
        }
        """
    ]


class _MessageConverterRegistry:
  """Message converter registry."""

  def __init__(self):
    self._name_to_converter: dict[str, Type[MessageConverter]] = {}
    self._type_to_converters: dict[Type[Any], list[Type[MessageConverter]]] = (
        collections.defaultdict(list)
    )

  def register(self, converter: Type['MessageConverter']) -> None:
    """Registers a message converter."""
    self._name_to_converter[converter.FORMAT_ID] = converter
    if converter.OUTPUT_TYPE is not None:
      self._type_to_converters[converter.OUTPUT_TYPE].append(converter)

  def get_by_type(self, t: Type[Any], **kwargs) -> 'MessageConverter':
    """Returns a message converter for the given type."""
    t = self._type_to_converters[t]
    if not t:
      raise TypeError(
          f'Cannot convert Message to {t!r}.'
      )
    if len(t) > 1:
      raise TypeError(
          f'More than one converters found for output type {t!r}. '
          f'Please specify one for this conversion: {[x.FORMAT_ID for x in t]}.'
      )
    return t[0](**kwargs)

  def get_by_format(self, format: str, **kwargs) -> 'MessageConverter':   # pylint: disable=redefined-builtin
    """Returns a message converter for the given format."""
    if format not in self._name_to_converter:
      raise ValueError(f'Unsupported format: {format!r}.')
    return self._name_to_converter[format](**kwargs)

  def get(
      self,
      format_or_type: str | Type[Any], **kwargs
  ) -> 'MessageConverter':
    """Returns a message converter for the given format or type."""
    if isinstance(format_or_type, str):
      return self.get_by_format(format_or_type, **kwargs)
    assert isinstance(format_or_type, type), format_or_type
    return self.get_by_type(format_or_type, **kwargs)

  def convertible_formats(self) -> list[str]:
    """Returns a list of converter names."""
    return sorted(list(self._name_to_converter.keys()))

  def convertible_types(self) -> list[Type[Any]]:
    """Returns a list of converter types."""
    return list(self._type_to_converters.keys())


class MessageConverter(pg.Object):
  """Interface for converting a Langfun message to other formats."""

  # A global unique identifier for the converter.
  FORMAT_ID: ClassVar[str]

  # The output type of the converter.
  # If None, the converter will not be registered to handle
  # `lf.Message.to_value(output_type)`.
  OUTPUT_TYPE: ClassVar[Type[Any] | None] = None

  _REGISTRY = _MessageConverterRegistry()

  def __init_subclass__(cls, *args, **kwargs):
    super().__init_subclass__(*args, **kwargs)
    if not inspect.isabstract(cls):
      cls._REGISTRY.register(cls)

  @abc.abstractmethod
  def to_value(self, message: Message) -> Any:
    """Converts a Langfun message to other formats."""

  @abc.abstractmethod
  def from_value(self, value: Any) -> Message:
    """Returns a Langfun message from other formats."""

  @classmethod
  def _safe_read(
      cls,
      data: Mapping[str, Any],
      key: str,
      default: Any = pg.MISSING_VALUE
  ) -> Any:
    """Safe reads a key from a mapping."""
    if not isinstance(data, Mapping):
      raise ValueError(f'Invalid data type: {data!r}.')
    if key not in data:
      if pg.MISSING_VALUE == default:
        raise ValueError(f'Missing key {key!r} in {data!r}')
      return default
    return data[key]

  @classmethod
  def get_role(cls, message: Message) -> str:
    """Returns the role of the message."""
    if isinstance(message, SystemMessage):
      return 'system'
    elif isinstance(message, UserMessage):
      return 'user'
    elif isinstance(message, AIMessage):
      return 'assistant'
    else:
      raise ValueError(f'Unsupported message type: {message!r}.')

  @classmethod
  def get_message_cls(cls, role: str) -> type[Message]:
    """Returns the message class of the message."""
    match role:
      case 'system':
        return SystemMessage
      case 'user':
        return UserMessage
      case 'assistant':
        return AIMessage
      case _:
        raise ValueError(f'Unsupported role: {role!r}.')

  @classmethod
  def get(cls, format_or_type: str | Type[Any], **kwargs) -> 'MessageConverter':
    """Returns a message converter."""
    return cls._REGISTRY.get(format_or_type, **kwargs)

  @classmethod
  def get_by_format(cls, format: str, **kwargs) -> 'MessageConverter':  # pylint: disable=redefined-builtin
    """Returns a message converter for the given format."""
    return cls._REGISTRY.get_by_format(format, **kwargs)

  @classmethod
  def get_by_type(cls, t: Type[Any], **kwargs) -> 'MessageConverter':
    """Returns a message converter for the given type."""
    return cls._REGISTRY.get_by_type(t, **kwargs)

  @classmethod
  def convertible_formats(cls) -> list[str]:
    """Returns a list of converter names."""
    return cls._REGISTRY.convertible_formats()

  @classmethod
  def convertible_types(cls) -> list[Type[Any]]:
    """Returns a list of converter types."""
    return cls._REGISTRY.convertible_types()

#
# Messages of different roles.
#


@pg.use_init_args(['text', 'sender', 'metadata'])
class UserMessage(Message):
  """Message sent from a human user."""

  sender = 'User'


@pg.use_init_args(['text', 'sender', 'metadata'])
class AIMessage(Message):
  """Message sent from an agent."""

  sender = 'AI'


@pg.use_init_args(['text', 'sender', 'metadata'])
class SystemMessage(Message):
  """Message sent from the system or environment."""

  sender = 'System'


@pg.use_init_args(['text', 'sender', 'metadata'])
class MemoryRecord(Message):
  """Message used as a memory record."""

  sender = 'Memory'



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Audio tests."""
import unittest
from unittest import mock

from langfun.core.modalities import audio as audio_lib
import pyglove as pg


content_bytes = (
    b'RIFF$\x00\x00\x00WAVEfmt'
    b' \x10\x00\x00\x00\x01\x00\x01\x00\x00\x04\x00\x00\x00\x04\x00\x00\x01\x00\x08\x00data\x00\x00\x00\x00'
)


def mock_request(*args, **kwargs):
  del args, kwargs
  return pg.Dict(content=content_bytes)


class AudioTest(unittest.TestCase):

  def test_audio_content(self):
    audio = audio_lib.Audio.from_bytes(content_bytes)
    self.assertIn(audio.mime_type, ('audio/wave', 'audio/x-wav'))
    self.assertIn(audio.audio_format, ('wave', 'x-wav'))
    self.assertEqual(audio.to_bytes(), content_bytes)

  def test_bad_audio(self):
    audio = audio_lib.Audio.from_bytes(b'bad')
    with self.assertRaisesRegex(ValueError, 'Expected MIME type'):
      _ = audio.audio_format


class AudioFileTest(unittest.TestCase):

  def test_audio_file(self):
    audio = audio_lib.Audio.from_uri('http://mock/web/a.wav')
    with mock.patch('requests.get') as mock_requests_get:
      mock_requests_get.side_effect = mock_request
      self.assertIn(audio.audio_format, ('wave', 'x-wav'))
      self.assertIn(audio.mime_type, ('audio/wave', 'audio/x-wav'))
      self.assertEqual(
          audio._raw_html(),
          '<audio controls> <source src="http://mock/web/a.wav"> </audio>',
      )
      self.assertEqual(audio.to_bytes(), content_bytes)


if __name__ == '__main__':
  unittest.main()



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Gemini REST API (Shared by Google GenAI and Vertex AI)."""

import datetime
import functools
from typing import Annotated, Any

import langfun.core as lf
from langfun.core import modalities as lf_modalities
from langfun.core.data.conversion import gemini as gemini_conversion  # pylint: disable=unused-import
from langfun.core.llms import rest
import pyglove as pg


class GeminiModelInfo(lf.ModelInfo):
  """Gemini model info."""

  # Constants for supported MIME types.
  INPUT_IMAGE_TYPES = [
      'image/png',
      'image/jpeg',
      'image/webp',
      'image/heic',
      'image/heif',
  ]

  INPUT_AUDIO_TYPES = [
      'audio/aac',
      'audio/flac',
      'audio/mp3',
      'audio/m4a',
      'audio/mpeg',
      'audio/mpga',
      'audio/mp4',
      'audio/opus',
      'audio/pcm',
      'audio/wav',
      'audio/webm',
  ]

  INPUT_VIDEO_TYPES = [
      'video/mov',
      'video/mpeg',
      'video/mpegps',
      'video/mpg',
      'video/mp4',
      'video/webm',
      'video/wmv',
      'video/x-flv',
      'video/3gpp',
      'video/quicktime',
  ]

  INPUT_DOC_TYPES = [
      'application/pdf',
      'text/plain',
      'text/csv',
      'text/html',
      'text/xml',
      'text/x-script.python',
      'application/json',
  ]

  ALL_SUPPORTED_INPUT_TYPES = (
      INPUT_IMAGE_TYPES
      + INPUT_AUDIO_TYPES
      + INPUT_VIDEO_TYPES
      + INPUT_DOC_TYPES
  )

  LINKS = dict(
      models='https://ai.google.dev/gemini-api/docs/models/gemini',
      pricing='https://ai.google.dev/gemini-api/docs/pricing',
      rate_limits='https://ai.google.dev/gemini-api/docs/models/gemini',
      error_codes='https://ai.google.dev/gemini-api/docs/troubleshooting?lang=python#error-codes',
  )

  class Pricing(lf.ModelInfo.Pricing):
    """Pricing for Gemini models."""

    cost_per_1m_cached_input_tokens_with_prompt_longer_than_128k: Annotated[
        float | None,
        (
            'The cost per 1M cached input tokens for prompts longer than 128K. '
            'If None, the 128k constraint is not applicable.'
        )
    ] = None

    cost_per_1m_input_tokens_with_prompt_longer_than_128k: Annotated[
        float | None,
        (
            'The cost per 1M input tokens for prompts longer than 128K. '
            'If None, the 128k constraint is not applicable.'
        )
    ] = None

    cost_per_1m_output_tokens_with_prompt_longer_than_128k: Annotated[
        float | None,
        (
            'The cost per 1M output tokens for prompts longer than 128K.'
            'If None, the 128k constraint is not applicable.'
        )
    ] = None

    def estimate_cost(self, usage: lf.LMSamplingUsage) -> float | None:
      """Estimates the cost of using the model. Subclass could override.

      Args:
        usage: The usage information of the model.

      Returns:
        The estimated cost in US dollars. If None, cost estimating is not
        supported on the model.
      """
      if (usage.prompt_tokens is None
          or usage.prompt_tokens < 128_000
          or not self.cost_per_1m_input_tokens_with_prompt_longer_than_128k):
        return super().estimate_cost(usage)

      return (
          self.cost_per_1m_input_tokens_with_prompt_longer_than_128k
          * usage.prompt_tokens
          + self.cost_per_1m_output_tokens_with_prompt_longer_than_128k
          * (usage.total_tokens - usage.prompt_tokens)
      ) / 1000_000

  experimental: Annotated[
      bool,
      (
          'If True, the model is experimental and may retire without notice.'
      )
  ] = False


# !!! PLEASE KEEP MODELS SORTED BY MODEL FAMILY AND RELEASE DATE !!!


SUPPORTED_MODELS = [
    #
    # Production models.
    #
    # Gemini 2.5 Flash
    GeminiModelInfo(
        model_id='gemini-2.5-flash',
        in_service=True,
        provider=pg.oneof(['Google GenAI', 'VertexAI']),
        model_type='instruction-tuned',
        description='Gemini 2.5 Flash GA.',
        release_date=datetime.datetime(2025, 6, 17),
        input_modalities=GeminiModelInfo.ALL_SUPPORTED_INPUT_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=1_048_576,
            max_output_tokens=65_536,
        ),
        pricing=GeminiModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=0.3,
            cost_per_1m_input_tokens=0.3,
            cost_per_1m_output_tokens=2.5,
        ),
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=2000,
            max_tokens_per_minute=4_000_000,
        ),
    ),
    # Gemini 2.5 Pro
    GeminiModelInfo(
        model_id='gemini-2.5-pro',
        in_service=True,
        provider=pg.oneof(['Google GenAI', 'VertexAI']),
        model_type='instruction-tuned',
        description='Gemini 2.5 Pro GA.',
        release_date=datetime.datetime(2025, 6, 17),
        input_modalities=GeminiModelInfo.ALL_SUPPORTED_INPUT_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=1_048_576,
            max_output_tokens=65_536,
        ),
        pricing=GeminiModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=1.25,
            cost_per_1m_input_tokens=1.25,
            cost_per_1m_output_tokens=10.0,
        ),
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=2000,
            max_tokens_per_minute=4_000_000,
        ),
    ),
    # Gemini 2.5 Pro 0605
    GeminiModelInfo(
        model_id='gemini-2.5-pro-preview-06-05',
        in_service=True,
        provider=pg.oneof(['Google GenAI', 'VertexAI']),
        model_type='instruction-tuned',
        description='Gemini 2.5 Pro.',
        release_date=datetime.datetime(2025, 6, 5),
        input_modalities=GeminiModelInfo.ALL_SUPPORTED_INPUT_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=1_048_576,
            max_output_tokens=65_536,
        ),
        pricing=GeminiModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=1.25,
            cost_per_1m_input_tokens=1.25,
            cost_per_1m_output_tokens=10.0,
        ),
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=2000,
            max_tokens_per_minute=4_000_000,
        ),
    ),
    # Gemini 2.5 Flash Preview 0520
    GeminiModelInfo(
        model_id='gemini-2.5-flash-preview-05-20',
        in_service=True,
        provider=pg.oneof(['Google GenAI', 'VertexAI']),
        model_type='instruction-tuned',
        description='Gemini 2.5 Flash.',
        release_date=datetime.datetime(2025, 5, 20),
        input_modalities=GeminiModelInfo.ALL_SUPPORTED_INPUT_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=1_048_576,
            max_output_tokens=65_536,
        ),
        pricing=GeminiModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=0.15,
            cost_per_1m_input_tokens=0.15,
            cost_per_1m_output_tokens=0.6,
        ),
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=2000,
            max_tokens_per_minute=4_000_000,
        ),
    ),
    # Gemini 2.5 Pro Preview
    GeminiModelInfo(
        model_id='gemini-2.5-pro-preview-05-06',
        in_service=True,
        provider=pg.oneof(['Google GenAI', 'VertexAI']),
        model_type='instruction-tuned',
        description='Gemini 2.5 Pro.',
        release_date=datetime.datetime(2025, 5, 6),
        input_modalities=GeminiModelInfo.ALL_SUPPORTED_INPUT_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=1_048_576,
            max_output_tokens=65_536,
        ),
        pricing=GeminiModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=1.25,
            cost_per_1m_input_tokens=1.25,
            cost_per_1m_output_tokens=10.0,
        ),
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=2000,
            max_tokens_per_minute=4_000_000,
        ),
    ),
    # Gemini 2.5 Flash Preview
    GeminiModelInfo(
        model_id='gemini-2.5-flash-preview-04-17',
        in_service=True,
        provider=pg.oneof(['Google GenAI', 'VertexAI']),
        model_type='instruction-tuned',
        description='Gemini 2.5 Flash.',
        release_date=datetime.datetime(2025, 4, 17),
        input_modalities=GeminiModelInfo.ALL_SUPPORTED_INPUT_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=1_048_576,
            max_output_tokens=65_536,
        ),
        pricing=GeminiModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=0.15,
            cost_per_1m_input_tokens=0.15,
            cost_per_1m_output_tokens=0.6,
        ),
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=2000,
            max_tokens_per_minute=4_000_000,
        ),
    ),
    # Gemini 2.5 Pro Preview
    GeminiModelInfo(
        model_id='gemini-2.5-pro-preview-03-25',
        in_service=True,
        provider=pg.oneof(['Google GenAI', 'VertexAI']),
        model_type='instruction-tuned',
        description='Gemini 2.5 Pro.',
        release_date=datetime.datetime(2025, 3, 25),
        input_modalities=GeminiModelInfo.ALL_SUPPORTED_INPUT_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=1_048_576,
            max_output_tokens=65_536,
        ),
        pricing=GeminiModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=1.25,
            cost_per_1m_input_tokens=1.25,
            cost_per_1m_output_tokens=10.0,
        ),
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=2000,
            max_tokens_per_minute=4_000_000,
        ),
    ),
    GeminiModelInfo(
        model_id='gemini-2.5-pro-exp-03-25',
        in_service=True,
        provider=pg.oneof(['VertexAI']),
        model_type='instruction-tuned',
        description='Gemini 2.5 Pro.',
        release_date=datetime.datetime(2025, 3, 25),
        input_modalities=GeminiModelInfo.ALL_SUPPORTED_INPUT_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=1_048_576,
            max_output_tokens=65_536,
        ),
        pricing=GeminiModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=1.25,
            cost_per_1m_input_tokens=1.25,
            cost_per_1m_output_tokens=10.0,
        ),
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=20,
            max_tokens_per_minute=1_000_000,
        ),
    ),
    # Gemini 2.0 Flash.
    GeminiModelInfo(
        model_id='gemini-2.0-flash',
        in_service=True,
        provider=pg.oneof(['Google GenAI', 'VertexAI']),
        model_type='instruction-tuned',
        description='Gemini 2.0 Flash model.',
        release_date=datetime.datetime(2025, 2, 5),
        input_modalities=GeminiModelInfo.ALL_SUPPORTED_INPUT_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=1_048_576,
            max_output_tokens=8_192,
        ),
        pricing=GeminiModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=0.025,
            cost_per_1m_input_tokens=0.1,
            cost_per_1m_output_tokens=0.4,
        ),
        rate_limits=lf.ModelInfo.RateLimits(
            # Tier 4 rate limits
            max_requests_per_minute=2000,
            max_tokens_per_minute=4_000_000,
        ),
    ),
    GeminiModelInfo(
        model_id='gemini-2.0-flash-001',
        in_service=True,
        provider=pg.oneof(['Google GenAI', 'VertexAI']),
        model_type='instruction-tuned',
        description='Gemini 2.0 Flash model (version 001).',
        release_date=datetime.datetime(2025, 2, 5),
        input_modalities=GeminiModelInfo.ALL_SUPPORTED_INPUT_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=1_048_576,
            max_output_tokens=8_192,
        ),
        pricing=GeminiModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=0.025,
            cost_per_1m_input_tokens=0.1,
            cost_per_1m_output_tokens=0.4,
        ),
        rate_limits=lf.ModelInfo.RateLimits(
            # Tier 4 rate limits
            max_requests_per_minute=2000,
            max_tokens_per_minute=4_000_000,
        ),
    ),
    # Gemini 2.0 Flash Lite.
    GeminiModelInfo(
        model_id='gemini-2.0-flash-lite-preview-02-05',
        in_service=True,
        provider=pg.oneof(['Google GenAI', 'VertexAI']),
        model_type='instruction-tuned',
        description='Gemini 2.0 Lite preview model.',
        release_date=datetime.datetime(2025, 2, 5),
        input_modalities=GeminiModelInfo.ALL_SUPPORTED_INPUT_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=1_048_576,
            max_output_tokens=8_192,
        ),
        pricing=GeminiModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=0.01875,
            cost_per_1m_input_tokens=0.075,
            cost_per_1m_output_tokens=0.3,
        ),
        rate_limits=lf.ModelInfo.RateLimits(
            # Tier 4 rate limits
            max_requests_per_minute=4000,
            max_tokens_per_minute=4_000_000,
        ),
    ),
    # Gemini 1.5 Flash.
    GeminiModelInfo(
        model_id='gemini-1.5-flash',
        alias_for='gemini-1.5-flash-002',
        provider=pg.oneof(['Google GenAI', 'VertexAI']),
        in_service=True,
        model_type='instruction-tuned',
        description='Gemini 1.5 Flash model (latest stable).',
        release_date=datetime.datetime(2024, 9, 30),
        input_modalities=GeminiModelInfo.ALL_SUPPORTED_INPUT_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=1_048_576,
            max_output_tokens=8_192,
        ),
        pricing=GeminiModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=0.01875,
            cost_per_1m_input_tokens=0.075,
            cost_per_1m_output_tokens=0.3,
            cost_per_1m_cached_input_tokens_with_prompt_longer_than_128k=0.0375,
            cost_per_1m_input_tokens_with_prompt_longer_than_128k=0.15,
            cost_per_1m_output_tokens_with_prompt_longer_than_128k=0.6,
        ),
        rate_limits=lf.ModelInfo.RateLimits(
            # Tier 4 rate limits
            max_requests_per_minute=2000,
            max_tokens_per_minute=4_000_000,
        ),
    ),
    GeminiModelInfo(
        model_id='gemini-1.5-flash-001',
        in_service=True,
        provider=pg.oneof(['Google GenAI', 'VertexAI']),
        model_type='instruction-tuned',
        description='Gemini 1.5 Flash model (version 001).',
        input_modalities=GeminiModelInfo.ALL_SUPPORTED_INPUT_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=1_048_576,
            max_output_tokens=8_192,
        ),
        pricing=GeminiModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=0.01875,
            cost_per_1m_input_tokens=0.075,
            cost_per_1m_output_tokens=0.3,
            cost_per_1m_cached_input_tokens_with_prompt_longer_than_128k=0.0375,
            cost_per_1m_input_tokens_with_prompt_longer_than_128k=0.15,
            cost_per_1m_output_tokens_with_prompt_longer_than_128k=0.6,
        ),
        rate_limits=lf.ModelInfo.RateLimits(
            # Tier 4 rate limits
            max_requests_per_minute=2000,
            max_tokens_per_minute=4_000_000,
        ),
    ),
    GeminiModelInfo(
        model_id='gemini-1.5-flash-002',
        in_service=True,
        provider=pg.oneof(['Google GenAI', 'VertexAI']),
        model_type='instruction-tuned',
        description='Gemini 1.5 Flash model (version 002).',
        input_modalities=GeminiModelInfo.ALL_SUPPORTED_INPUT_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=1_048_576,
            max_output_tokens=8_192,
        ),
        pricing=GeminiModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=0.01875,
            cost_per_1m_input_tokens=0.075,
            cost_per_1m_output_tokens=0.3,
            cost_per_1m_cached_input_tokens_with_prompt_longer_than_128k=0.0375,
            cost_per_1m_input_tokens_with_prompt_longer_than_128k=0.15,
            cost_per_1m_output_tokens_with_prompt_longer_than_128k=0.6,
        ),
        rate_limits=lf.ModelInfo.RateLimits(
            # Tier 4 rate limits
            max_requests_per_minute=2000,
            max_tokens_per_minute=4_000_000,
        ),
    ),
    # Gemini 1.5 Flash-8B.
    GeminiModelInfo(
        model_id='gemini-1.5-flash-8b',
        in_service=True,
        provider='Google GenAI',
        model_type='instruction-tuned',
        description='Gemini 1.5 Flash 8B model (latest stable).',
        input_modalities=GeminiModelInfo.ALL_SUPPORTED_INPUT_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=1_048_576,
            max_output_tokens=8_192,
        ),
        pricing=GeminiModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=0.01,
            cost_per_1m_input_tokens=0.0375,
            cost_per_1m_output_tokens=0.15,
            cost_per_1m_cached_input_tokens_with_prompt_longer_than_128k=0.02,
            cost_per_1m_input_tokens_with_prompt_longer_than_128k=0.075,
            cost_per_1m_output_tokens_with_prompt_longer_than_128k=0.3,
        ),
        rate_limits=lf.ModelInfo.RateLimits(
            # Tier 4 rate limits
            max_requests_per_minute=4000,
            max_tokens_per_minute=4_000_000,
        ),
    ),
    GeminiModelInfo(
        model_id='gemini-1.5-flash-8b-001',
        in_service=True,
        provider='Google GenAI',
        model_type='instruction-tuned',
        description='Gemini 1.5 Flash 8B model (version 001).',
        input_modalities=GeminiModelInfo.ALL_SUPPORTED_INPUT_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=1_048_576,
            max_output_tokens=8_192,
        ),
        pricing=GeminiModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=0.01,
            cost_per_1m_input_tokens=0.0375,
            cost_per_1m_output_tokens=0.15,
            cost_per_1m_cached_input_tokens_with_prompt_longer_than_128k=0.02,
            cost_per_1m_input_tokens_with_prompt_longer_than_128k=0.075,
            cost_per_1m_output_tokens_with_prompt_longer_than_128k=0.3,
        ),
        rate_limits=lf.ModelInfo.RateLimits(
            # Tier 4 rate limits
            max_requests_per_minute=4000,
            max_tokens_per_minute=4_000_000,
        ),
    ),
    # Gemini 1.5 Pro.
    GeminiModelInfo(
        model_id='gemini-1.5-pro',
        alias_for='gemini-1.5-pro-002',
        in_service=True,
        provider=pg.oneof(['Google GenAI', 'VertexAI']),
        model_type='instruction-tuned',
        description='Gemini 1.5 Pro model (latest stable).',
        input_modalities=GeminiModelInfo.ALL_SUPPORTED_INPUT_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=2_097_152,
            max_output_tokens=8_192,
        ),
        pricing=GeminiModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=0.3125,
            cost_per_1m_input_tokens=1.25,
            cost_per_1m_output_tokens=5,
            cost_per_1m_cached_input_tokens_with_prompt_longer_than_128k=0.625,
            cost_per_1m_input_tokens_with_prompt_longer_than_128k=2.5,
            cost_per_1m_output_tokens_with_prompt_longer_than_128k=10,
        ),
        rate_limits=lf.ModelInfo.RateLimits(
            # Tier 4 rate limits
            max_requests_per_minute=1000,
            max_tokens_per_minute=4_000_000,
        ),
    ),
    GeminiModelInfo(
        model_id='gemini-1.5-pro-001',
        in_service=True,
        provider=pg.oneof(['Google GenAI', 'VertexAI']),
        model_type='instruction-tuned',
        description='Gemini 1.5 Pro model (version 001).',
        input_modalities=GeminiModelInfo.ALL_SUPPORTED_INPUT_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=2_097_152,
            max_output_tokens=8_192,
        ),
        pricing=GeminiModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=0.3125,
            cost_per_1m_input_tokens=1.25,
            cost_per_1m_output_tokens=5,
            cost_per_1m_cached_input_tokens_with_prompt_longer_than_128k=0.625,
            cost_per_1m_input_tokens_with_prompt_longer_than_128k=2.5,
            cost_per_1m_output_tokens_with_prompt_longer_than_128k=10,
        ),
        rate_limits=lf.ModelInfo.RateLimits(
            # Tier 4 rate limits
            max_requests_per_minute=1000,
            max_tokens_per_minute=4_000_000,
        ),
    ),
    GeminiModelInfo(
        model_id='gemini-1.5-pro-002',
        in_service=True,
        provider=pg.oneof(['Google GenAI', 'VertexAI']),
        model_type='instruction-tuned',
        description='Gemini 1.5 Pro model (version 002).',
        input_modalities=GeminiModelInfo.ALL_SUPPORTED_INPUT_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=2_097_152,
            max_output_tokens=8_192,
        ),
        pricing=GeminiModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=0.3125,
            cost_per_1m_input_tokens=1.25,
            cost_per_1m_output_tokens=5,
            cost_per_1m_cached_input_tokens_with_prompt_longer_than_128k=0.625,
            cost_per_1m_input_tokens_with_prompt_longer_than_128k=2.5,
            cost_per_1m_output_tokens_with_prompt_longer_than_128k=10,
        ),
        rate_limits=lf.ModelInfo.RateLimits(
            # Tier 4 rate limits
            max_requests_per_minute=1000,
            max_tokens_per_minute=4_000_000,
        ),
    ),
    #
    # Experimental models.
    #
    GeminiModelInfo(
        model_id='gemini-2.5-flash-image-preview',
        in_service=True,
        experimental=True,
        provider=pg.oneof(['Google GenAI', 'VertexAI']),
        model_type='instruction-tuned',
        description='Gemini 2.5 Flash Image Preview model.',
        release_date=datetime.datetime(2025, 8, 17),
        input_modalities=GeminiModelInfo.INPUT_IMAGE_TYPES
        + GeminiModelInfo.INPUT_DOC_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=32_768,
            max_output_tokens=32_768,
        ),
    ),
    GeminiModelInfo(
        model_id='gemini-2.0-pro-exp-02-05',
        in_service=True,
        experimental=True,
        provider=pg.oneof(['Google GenAI', 'VertexAI']),
        model_type='instruction-tuned',
        description='Gemini 2.0 Pro experimental model (02/05/2025).',
        release_date=datetime.datetime(2025, 2, 5),
        input_modalities=GeminiModelInfo.ALL_SUPPORTED_INPUT_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=1_048_576,
            max_output_tokens=8_192,
        ),
    ),
    GeminiModelInfo(
        model_id='gemini-2.0-flash-thinking-exp-01-21',
        in_service=True,
        experimental=True,
        provider=pg.oneof(['Google GenAI', 'VertexAI']),
        model_type='thinking',
        description=(
            'Gemini 2.0 Flash thinking experimental model (01/21/2025).'
        ),
        release_date=datetime.datetime(2025, 1, 21),
        input_modalities=GeminiModelInfo.INPUT_IMAGE_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=1_048_576,
            max_output_tokens=8_192,
        ),
    ),
    GeminiModelInfo(
        model_id='gemini-exp-1206',
        in_service=True,
        experimental=True,
        provider='Google GenAI',
        model_type='instruction-tuned',
        description='Gemini year 1 experimental model (12/06/2024)',
        release_date=datetime.datetime(2025, 1, 21),
        input_modalities=GeminiModelInfo.ALL_SUPPORTED_INPUT_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=1_048_576,
            max_output_tokens=8_192,
        ),
    ),
    GeminiModelInfo(
        model_id='learnlm-1.5-pro-experimental',
        in_service=True,
        experimental=True,
        provider='Google GenAI',
        model_type='instruction-tuned',
        description='Gemini experimental model on learning science principles.',
        url='https://ai.google.dev/gemini-api/docs/learnlm',
        release_date=datetime.datetime(2024, 11, 19),
        input_modalities=GeminiModelInfo.ALL_SUPPORTED_INPUT_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=1_048_576,
            max_output_tokens=8_192,
        ),
    ),
]


_SUPPORTED_MODELS_BY_ID = {m.model_id: m for m in SUPPORTED_MODELS}


@pg.use_init_args(['model'])
class Gemini(rest.REST):
  """Language models provided by Google GenAI."""

  model: pg.typing.Annotated[
      pg.typing.Enum(
          pg.MISSING_VALUE, [m.model_id for m in SUPPORTED_MODELS]
      ),
      'The name of the model to use.',
  ]

  response_modalities: pg.typing.Annotated[
      list[str] | None,
      'Response modalities. It is needed for models whose response modalities '
      + 'are more than plain text.',
  ] = None

  @functools.cached_property
  def model_info(self) -> GeminiModelInfo:
    return _SUPPORTED_MODELS_BY_ID[self.model]

  @classmethod
  def dir(cls):
    return [m.model_id for m in SUPPORTED_MODELS if m.in_service]

  @property
  def headers(self):
    return {
        'Content-Type': 'application/json; charset=utf-8',
    }

  def request(
      self, prompt: lf.Message, sampling_options: lf.LMSamplingOptions
  ) -> dict[str, Any]:
    request = dict(
        generationConfig=self._generation_config(prompt, sampling_options)
    )
    def modality_conversion(chunk: str | lf.Modality) -> Any:
      if isinstance(chunk, lf_modalities.Mime):
        try:
          return chunk.make_compatible(
              self.model_info.input_modalities + ['text/plain']
          )
        except lf.ModalityError as e:
          raise lf.ModalityError(f'Unsupported modality: {chunk!r}') from e
      return chunk

    if system_message := prompt.get('system_message'):
      assert isinstance(system_message, lf.SystemMessage), type(system_message)
      request['systemInstruction'] = system_message.as_format(
          'gemini', chunk_preprocessor=modality_conversion
      )

    contents = []
    contents.append(
        prompt.as_format('gemini', chunk_preprocessor=modality_conversion)
    )
    request['contents'] = contents
    return request

  def _generation_config(
      self, prompt: lf.Message, options: lf.LMSamplingOptions
  ) -> dict[str, Any]:
    """Returns a dict as generation config for prompt and LMSamplingOptions."""
    config = dict(
        temperature=options.temperature,
        maxOutputTokens=options.max_tokens,
        candidateCount=options.n,
        topK=options.top_k,
        topP=options.top_p,
        stopSequences=options.stop,
        seed=options.random_seed,
        responseLogprobs=options.logprobs,
        logprobs=options.top_logprobs,
    )

    if json_schema := prompt.metadata.get('json_schema'):
      if not isinstance(json_schema, dict):
        raise ValueError(
            f'`json_schema` must be a dict, got {json_schema!r}.'
        )
      json_schema = pg.to_json(json_schema)
      config['responseSchema'] = json_schema
      config['responseMimeType'] = 'application/json'
      prompt.metadata.formatted_text = (
          prompt.text
          + '\n\n [RESPONSE FORMAT (not part of prompt)]\n'
          + pg.to_json_str(json_schema, json_indent=2)
      )
    if options.max_thinking_tokens is not None:
      config['thinkingConfig'] = {
          'includeThoughts': options.max_thinking_tokens > 0,
          'thinkingBudget': options.max_thinking_tokens,
      }

    if self.response_modalities:
      config['responseModalities'] = self.response_modalities
      if 'IMAGE' in self.response_modalities:
        config.pop('responseLogprobs', None)
        config.pop('logprobs', None)
    return config

  def result(self, json: dict[str, Any]) -> lf.LMSamplingResult:
    candidates = json.get('candidates')
    if candidates is None:
      raise lf.TemporaryLMError(
          'No candidates found in response. This is a Gemini API issue that '
          'happens occasionally, and retrying should fix it. '
      )
    messages = [
        lf.Message.from_value(candidate['content'], format='gemini')
        for candidate in candidates
    ]
    usage = json['usageMetadata']
    input_tokens = usage['promptTokenCount']
    # NOTE(daiyip): We saw cases that `candidatesTokenCount` is not present.
    # Therefore, we use 0 as the default value.
    output_tokens = usage.get('candidatesTokenCount', 0)
    thinking_tokens = usage.get('thoughtsTokenCount', 0)
    total_tokens = usage.get('totalTokenCount', 0)

    return lf.LMSamplingResult(
        [lf.LMSample(message) for message in messages],
        usage=lf.LMSamplingUsage(
            prompt_tokens=input_tokens,
            completion_tokens=output_tokens,
            total_tokens=total_tokens,
            completion_tokens_details={
                'thinking_tokens': thinking_tokens,
            },
        ),
    )

  def _error(self, status_code: int, content: str) -> lf.LMError:
    if (
        status_code == 400
        and b'exceeds the maximum number of tokens' in content
    ):
      return lf.ContextLimitError(f'{status_code}: {content}')
    return super()._error(status_code, content)



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Symbolic description."""

import inspect
from typing import Any, Literal

import langfun.core as lf
from langfun.core.structured import mapping
import pyglove as pg


@pg.use_init_args(['examples'])
class _DescribeStructure(mapping.Mapping):
  """Describe a structured value in natural language."""

  input_title = 'PYTHON_OBJECT'
  context_title = 'CONTEXT_FOR_DESCRIPTION'
  output_title = 'NATURAL_LANGUAGE_TEXT'

  preamble = """
      Please help describe {{ input_title }} in natural language.

      INSTRUCTIONS:
        1. Do not add details which are not present in the object.
        2. If a field in the object has None as its value, do not mention it.
      """


def describe(
    value: Any,
    context: str | None = None,
    *,
    lm: lf.LanguageModel | None = None,
    examples: list[mapping.MappingExample] | None = None,
    cache_seed: int | None = 0,
    **kwargs,
) -> str:
  """Describes a structured value using natural language.

  Examples:

    ```
    class FlightDuration(pg.Object):
      hours: int
      minutes: int

    class Flight(pg.Object):
      airline: str
      flight_number: str
      departure_airport: str
      arrival_airport: str
      departure_time: str
      arrival_time: str
      duration: FlightDuration
      stops: int
      price: float

    text = lf.describe(
        Flight(
            airline='United Airlines',
            flight_number='UA2631',
            depature_airport: 'SFO',
            arrival_airport: 'JFK',
            depature_time: '2023-09-07T05:15:00',
            arrival_time: '2023-09-07T12:12:00',
            duration: FlightDuration(
                hours=7,
                minutes=57
            ),
            stops=1,
            price=227,
        ))
    print(text)

    >> The flight is operated by United Airlines, has the flight number UA2631,
    >> departs from San Francisco International Airport (SFO), arrives at John
    >> F. Kennedy International Airport (JFK), It departs at
    >> 2023-09-07T05:15:00, arrives at 2023-09-07T12:12:00, has a duration of 7
    >> hours and 57 minutes, makes 1 stop, and costs $227.
    ```

  Args:
    value: A structured value to be mapped.
    context: The context information for describing the structured value.
    lm: The language model to use. If not specified, the language model from
      `lf.context` context manager will be used.
    examples: An optional list of fewshot examples for helping parsing. If None,
      the default one-shot example will be added.
    cache_seed: Seed for computing cache key. The cache key is determined by a
      tuple of (lm, prompt, cache seed). If None, cache will be disabled for
      the query even cache is configured by the LM.
    **kwargs: Keyword arguments passed to the `lf.structured.DescribeStructure`.

  Returns:
    The parsed result based on the schema.
  """
  return _DescribeStructure(
      input=value,
      context=context,
      examples=examples or default_describe_examples(),
      **kwargs,
  )(lm=lm, cache_seed=cache_seed).text


def default_describe_examples() -> list[mapping.MappingExample]:
  """Default describe examples."""

  class Country(pg.Object):
    """A example dataclass for structured mapping."""

    name: str
    continents: list[
        Literal[
            'Africa',
            'Asia',
            'Europe',
            'Oceania',
            'North America',
            'South America',
        ]
    ]
    num_states: int
    neighbor_countries: list[str]
    population: int
    capital: str | None
    president: str | None

  return [
      mapping.MappingExample(
          context='Brief intro to United States',
          input=Country(
              name='The United States of America',
              continents=['North America'],
              num_states=50,
              neighbor_countries=[
                  'Canada',
                  'Mexico',
                  'Bahamas',
                  'Cuba',
                  'Russia',
              ],
              population=333000000,
              capital='Washington, D.C',
              president=None,
          ),
          output=inspect.cleandoc("""
              The United States of America is a country primarily located in North America
              consisting of fifty states. It shares land borders with Canada to its north
              and with Mexico to its south and has maritime borders with the Bahamas, Cuba,
              Russia, and other nations. With a population of over 333 million. The national
              capital of the United States is Washington, D.C.
              """),
      ),
  ]



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""langfun eval framework."""

# pylint: disable=g-importing-member
# pylint: disable=g-bad-import-order

from langfun.core.eval import v2

from langfun.core.eval.base import register
from langfun.core.eval.base import registered_names
from langfun.core.eval.base import get_evaluations
from langfun.core.eval.base import get
from langfun.core.eval.base import run

from langfun.core.eval.base import Evaluable
from langfun.core.eval.base import Evaluation
from langfun.core.eval.base import Suite
from langfun.core.eval.base import Summary

from langfun.core.eval.base import load
from langfun.core.eval.base import monitor
from langfun.core.eval.base import monitor_async

# Functors for loading inputs.
from langfun.core.eval.base import inputs_from
from langfun.core.eval.base import as_inputs

from langfun.core.eval.matching import Matching
from langfun.core.eval.scoring import Scoring

# Experiment patching.
from langfun.core.eval.patching import patch_member

from langfun.core.eval.patching import patch_lm
from langfun.core.eval.patching import patch_parsing_lm
from langfun.core.eval.patching import patch_inputs
from langfun.core.eval.patching import patch_prompt
from langfun.core.eval.patching import patch_schema_fn

# Placeholder for Google-internal imports.
# pylint: enable=g-bad-import-order
# pylint: enable=g-importing-member



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for Anthropic models."""

import base64
import os
from typing import Any
import unittest
from unittest import mock

import langfun.core as lf
from langfun.core import modalities as lf_modalities
from langfun.core.llms import anthropic
import pyglove as pg
import requests


def mock_requests_post(url: str, json: dict[str, Any], **kwargs):
  del url, kwargs

  response = requests.Response()
  response.status_code = 200

  # Construct base text from user/assistant messages payload
  messages_payload_text = '\n'.join(
      c['content'][0]['text']
      for c in json.get('messages', [])
      if c.get('content')
      and isinstance(c['content'], list)
      and c['content']
      and c['content'][0].get('type') == 'text'
      and 'text' in c['content'][0]
  )

  # Check for a system prompt in the request payload
  system_prompt_text = json.get('system')

  processed_text_parts = []
  if system_prompt_text:
    processed_text_parts.append(system_prompt_text)
  if messages_payload_text:
    processed_text_parts.append(messages_payload_text)

  processed_text = '\n'.join(processed_text_parts)

  response_content_text = (
      f'{processed_text} with temperature={json.get("temperature")}, '
      f'top_k={json.get("top_k")}, '
      f'top_p={json.get("top_p")}, '
      f'max_tokens={json.get("max_tokens")}, '
      f'stop={json.get("stop_sequences")}.'
  )

  response._content = pg.to_json_str({
      'content': [{'type': 'text', 'text': response_content_text}],
      'usage': {
          'input_tokens': (
              2
          ),  # Placeholder: adjust if tests need accurate token counts
          'output_tokens': (
              1
          ),  # Placeholder: adjust if tests need accurate token counts
      },
  }).encode()
  return response


image_content = (
    b'\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\x18\x00\x00\x00\x18\x04'
    b'\x03\x00\x00\x00\x12Y \xcb\x00\x00\x00\x18PLTE\x00\x00'
    b'\x00fff_chaag_cg_ch^ci_ciC\xedb\x94\x00\x00\x00\x08tRNS'
    b'\x00\n\x9f*\xd4\xff_\xf4\xe4\x8b\xf3a\x00\x00\x00>IDATx'
    b'\x01c \x05\x08)"\xd8\xcc\xae!\x06pNz\x88k\x19\\Q\xa8"\x10'
    b'\xc1\x14\x95\x01%\xc1\n\xa143Ta\xa8"D-\x84\x03QM\x98\xc3'
    b'\x1a\x1a\x1a@5\x0e\x04\xa0q\x88\x05\x00\x07\xf8\x18\xf9'
    b'\xdao\xd0|\x00\x00\x00\x00IEND\xaeB`\x82'
)

pdf_content = (
    b'%PDF-1.4\n1 0 obj\n<< /Type /Catalog /Pages 2 0 R >>\nendobj\n2 0 obj\n<<'
    b' /Type /Pages /Count 1 /Kids [3 0 R] >>\nendobj\n3 0 obj\n<< /Type /Page'
    b' /Parent 2 0 R /MediaBox [0 0 612 792] /Contents 4 0 R >>\nendobj\n4 0'
    b' obj\n<< /Length 44 >>\nstream\nBT /F1 24 Tf 100 700 Td (Hello, PDF'
    b' content!) Tj ET\nendstream\nendobj\n5 0 obj\n<< /Type /Font /Subtype'
    b' /Type1 /BaseFont /Helvetica >>\nendobj\nxref\n0 6\n0000000000 65535 f'
    b' \n0000000010 00000 n \n0000000079 00000 n \n0000000178 00000 n'
    b' \n0000000278 00000 n \n0000000407 00000 n \ntrailer\n<< /Size 6 /Root 1'
    b' 0 R >>\nstartxref\n517\n%%EOF'
)


def mock_mm_requests_post(url: str, json: dict[str, Any], **kwargs):
  del url, kwargs
  v = json['messages'][0]['content'][0]
  content = lf_modalities.Mime.from_bytes(base64.b64decode(v['source']['data']))

  response = requests.Response()
  response.status_code = 200
  response._content = pg.to_json_str({
      'content': [{
          'type': 'text',
          'text': f'{v["type"]}: {content.mime_type}',
      }],
      'usage': {
          'input_tokens': 2,
          'output_tokens': 1,
      },
  }).encode()
  return response


def mock_requests_post_error(status_code, error_type, error_message):
  def _mock_requests(url: str, json: dict[str, Any], **kwargs):
    del url, json, kwargs
    response = requests.Response()
    response.status_code = status_code
    response._content = pg.to_json_str(
        {
            'error': {
                'type': error_type,
                'message': error_message,
            }
        }
    ).encode()
    return response

  return _mock_requests


class AnthropicTest(unittest.TestCase):

  def test_basics(self):
    self.assertEqual(
        anthropic.Claude3Haiku().model_id, 'claude-3-haiku-20240307'
    )
    self.assertGreater(anthropic.Claude3Haiku().max_concurrency, 0)

  def test_model_alias(self):
    # Alias will be normalized to the official version.
    self.assertEqual(
        anthropic.Anthropic('claude-3-5-sonnet-20241022').model_id,
        'claude-3-5-sonnet-20241022'
    )
    self.assertEqual(
        anthropic.Anthropic('claude-3-5-sonnet-v2@20241022').model_id,
        'claude-3-5-sonnet-20241022'
    )

  def test_api_key(self):
    lm = anthropic.Claude3Haiku()
    with self.assertRaisesRegex(ValueError, 'Please specify `api_key`'):
      lm('hi')

    with mock.patch('requests.Session.post') as mock_request:
      mock_request.side_effect = mock_requests_post

      lm = anthropic.Claude3Haiku(api_key='fake key')
      self.assertRegex(lm('hello').text, 'hello.*')

      os.environ['ANTHROPIC_API_KEY'] = 'abc'
      lm = anthropic.Claude3Haiku()
      self.assertRegex(lm('hello').text, 'hello.*')
      del os.environ['ANTHROPIC_API_KEY']

  def test_call(self):
    with mock.patch('requests.Session.post') as mock_request:
      mock_request.side_effect = mock_requests_post
      lm = anthropic.Claude3Haiku(api_key='fake_key')
      response = lm('hello', temperature=0.0, top_k=0.1, top_p=0.2, stop=['\n'])
      self.assertEqual(
          response.text,
          (
              'hello with temperature=0.0, top_k=0.1, top_p=0.2, '
              "max_tokens=4096, stop=['\\n']."
          ),
      )
      self.assertIsNotNone(response.usage)
      self.assertIsNotNone(response.usage.prompt_tokens, 2)
      self.assertIsNotNone(response.usage.completion_tokens, 1)
      self.assertIsNotNone(response.usage.total_tokens, 3)
      self.assertGreater(response.usage.estimated_cost, 0)

  def test_call_with_system_message(self):
    with mock.patch('requests.Session.post') as mock_request:
      mock_request.side_effect = mock_requests_post
      lm = anthropic.Claude3Haiku(api_key='fake_key')
      response = lm(
          lf.UserMessage(
              'hello', system_message=lf.SystemMessage('system')
          ),
          temperature=0.0,
          top_k=0.1,
          top_p=0.2,
          stop=['\n'],
      )
      self.assertEqual(
          response.text,
          (
              'system\nhello with temperature=0.0, top_k=0.1, top_p=0.2, '
              "max_tokens=4096, stop=['\\n']."
          ),
      )

  def test_mm_call(self):
    with mock.patch('requests.Session.post') as mock_mm_request:
      mock_mm_request.side_effect = mock_mm_requests_post
      lm = anthropic.Claude3Haiku(api_key='fake_key')
      response = lm(lf_modalities.Image.from_bytes(image_content), lm=lm)
      self.assertEqual(response.text, 'image: image/png')

  def test_pdf_call(self):
    with mock.patch('requests.Session.post') as mock_mm_request:
      mock_mm_request.side_effect = mock_mm_requests_post
      lm = anthropic.Claude35Sonnet(api_key='fake_key')
      response = lm(lf_modalities.PDF.from_bytes(pdf_content), lm=lm)
      self.assertEqual(response.text, 'document: application/pdf')

  def test_call_errors(self):
    for status_code, error_type, error_message in [
        (429, 'rate_limit', 'Rate limit exceeded.'),
        (529, 'service_unavailable', 'Service unavailable.'),
        (500, 'bad_request', 'Bad request.'),
    ]:
      with mock.patch('requests.Session.post') as mock_mm_request:
        mock_mm_request.side_effect = mock_requests_post_error(
            status_code, error_type, error_message
        )
        lm = anthropic.Claude3Haiku(api_key='fake_key')
        with self.assertRaisesRegex(
            Exception, f'.*{status_code}: .*{error_message}'
        ):
          lm('hello', max_attempts=1)

  def test_call_with_context_limit_error(self):
    with mock.patch('requests.Session.post') as mock_request:
      mock_request.side_effect = mock_requests_post_error(
          413, 'bad_request', 'Prompt is too long.'
      )
      lm = anthropic.Claude3Haiku(api_key='fake_key')
      with self.assertRaisesRegex(
          lf.ContextLimitError, 'Prompt is too long.'
      ):
        lm('hello', max_attempts=1)

  def test_lm_get(self):
    self.assertIsInstance(
        lf.LanguageModel.get('claude-3-5-sonnet-latest'),
        anthropic.Anthropic,
    )


if __name__ == '__main__':
  unittest.main()
